# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import neutronclient.v2_0.client
from oslo_log import log as logging

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return NeutronDriver(name, keys, inbox, datapath, args)


class NeutronDriver(datasource_driver.PollingDataSourceDriver,
                    datasource_driver.ExecutionDriver):

    NETWORKS = "networks"
    NETWORKS_SUBNETS = "networks.subnets"
    PORTS = "ports"
    PORTS_ADDR_PAIRS = "ports.address_pairs"
    PORTS_SECURITY_GROUPS = "ports.security_groups"
    PORTS_BINDING_CAPABILITIES = "ports.binding_capabilities"
    PORTS_FIXED_IPS = "ports.fixed_ips"
    PORTS_FIXED_IPS_GROUPS = "ports.fixed_ips_groups"
    PORTS_EXTRA_DHCP_OPTS = "ports.extra_dhcp_opts"
    ROUTERS = "routers"
    ROUTERS_EXTERNAL_GATEWAYS = "routers.external_gateways"
    SECURITY_GROUPS = "security_groups"

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    networks_translator = {
        'translation-type': 'HDICT',
        'table-name': NETWORKS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'subnets', 'col': 'subnet_group_id',
              'translator': {'translation-type': 'LIST',
                             'table-name': 'networks.subnets',
                             'id-col': 'subnet_group_id',
                             'val-col': 'subnet',
                             'translator': value_trans}},
             {'fieldname': 'provider:physical_network',
              'translator': value_trans},
             {'fieldname': 'admin_state_up', 'translator': value_trans},
             {'fieldname': 'tenant_id', 'translator': value_trans},
             {'fieldname': 'provider:network_type', 'translator': value_trans},
             {'fieldname': 'router:external', 'translator': value_trans},
             {'fieldname': 'shared', 'translator': value_trans},
             {'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'provider:segmentation_id',
              'translator': value_trans})}

    ports_translator = {
        'translation-type': 'HDICT',
        'table-name': PORTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'allowed_address_pairs',
              'col': 'allowed_address_pairs_id',
              'translator': {'translation-type': 'LIST',
                             'table-name': PORTS_ADDR_PAIRS,
                             'id-col': 'allowed_address_pairs_id',
                             'val-col': 'address',
                             'translator': value_trans}},
             {'fieldname': 'security_groups',
              'col': 'security_groups_id',
              'translator': {'translation-type': 'LIST',
                             'table-name': PORTS_SECURITY_GROUPS,
                             'id-col': 'security_groups_id',
                             'val-col': 'security_group_id',
                             'translator': value_trans}},
             {'fieldname': 'extra_dhcp_opts',
              'col': 'extra_dhcp_opt_group_id',
              'translator': {'translation-type': 'LIST',
                             'table-name': PORTS_EXTRA_DHCP_OPTS,
                             'id-col': 'extra_dhcp_opt_group_id',
                             'val-col': 'dhcp_opt',
                             'translator': value_trans}},
             {'fieldname': 'binding:capabilities',
              'col': 'binding:capabilities_id',
              'translator': {'translation-type': 'VDICT',
                             'table-name': PORTS_BINDING_CAPABILITIES,
                             'id-col': 'binding:capabilities_id',
                             'key-col': 'key', 'val-col': 'value',
                             'translator': value_trans}},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'admin_state_up', 'translator': value_trans},
             {'fieldname': 'network_id', 'translator': value_trans},
             {'fieldname': 'tenant_id', 'translator': value_trans},
             {'fieldname': 'binding:vif_type', 'translator': value_trans},
             {'fieldname': 'device_owner', 'translator': value_trans},
             {'fieldname': 'mac_address', 'translator': value_trans},

             {'fieldname': 'fixed_ips',
              'col': 'fixed_ips',
              'translator': {'translation-type': 'LIST',
                             'table-name': PORTS_FIXED_IPS_GROUPS,
                             'id-col': 'fixed_ips_group_id',
                             'val-col': 'fixed_ip_id',
                             'translator': {'translation-type': 'VDICT',
                                            'table-name': PORTS_FIXED_IPS,
                                            'id-col': 'fixed_ip_id',
                                            'key-col': 'key',
                                            'val-col': 'value',
                                            'translator': value_trans}}},
             {'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'device_id', 'translator': value_trans},
             {'fieldname': 'binding:host_id', 'translator': value_trans})}

    routers_translator = {
        'translation-type': 'HDICT',
        'table-name': ROUTERS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'external_gateway_info',
              'translator': {'translation-type': 'VDICT',
                             'table-name': ROUTERS_EXTERNAL_GATEWAYS,
                             'id-col': 'external_gateway_info',
                             'key-col': 'key', 'val-col': 'value',
                             'translator': value_trans}},
             {'fieldname': 'networks', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'admin_state_up', 'translator': value_trans},
             {'fieldname': 'tenant_id', 'translator': value_trans},
             {'fieldname': 'id', 'translator': value_trans})}

    security_groups_translator = {
        'translation-type': 'HDICT',
        'table-name': SECURITY_GROUPS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'tenant_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'id', 'translator': value_trans})}

    TRANSLATORS = [networks_translator, ports_translator, routers_translator,
                   security_groups_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(NeutronDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = self.get_neutron_credentials(args)
        self.neutron = neutronclient.v2_0.client.Client(**self.creds)

        # Store raw state (result of API calls) so that we can
        #   avoid re-translating and re-sending if no changes occurred.
        #   Because translation is not deterministic (we're generating
        #   UUIDs), it's hard to tell if no changes occurred
        #   after performing the translation.
        self.raw_state = {}
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'neutron'
        result['description'] = ('Do not use this driver is deprecated')
        result['config'] = datasource_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def get_neutron_credentials(self, creds):
        d = {}
        d['username'] = creds['username']
        d['tenant_name'] = creds['tenant_name']
        d['password'] = creds['password']
        d['auth_url'] = creds['auth_url']
        return d

    def update_from_datasource(self):
        """Called when it is time to pull new data from this datasource.

        Sets self.state[tablename] = <set of tuples of strings/numbers>
        for every tablename exported by this datasource.
        """
        LOG.debug("Neutron grabbing networks")
        networks = self.neutron.list_networks()
        if ('networks' not in self.raw_state or
                networks != self.raw_state['networks']):
            self.raw_state['networks'] = networks
            self._translate_networks(networks)

        LOG.debug("Neutron grabbing ports")
        ports = self.neutron.list_ports()
        if 'ports' not in self.raw_state or ports != self.raw_state['ports']:
            self.raw_state['ports'] = ports
            self._translate_ports(ports)

        LOG.debug("Neutron grabbing routers")
        routers = self.neutron.list_routers()
        if ('routers' not in self.raw_state or
                routers != self.raw_state['routers']):
            self.raw_state['routers'] = routers
            self._translate_routers(routers)

        LOG.debug("Neutron grabbing security groups")
        security = self.neutron.list_security_groups()
        if ('security_groups' not in self.raw_state or
                security != self.raw_state['security_groups']):
            self.raw_state['security_groups'] = security
            self._translate_security_groups(security)

    def _translate_networks(self, obj):
        """Translate the networks represented by OBJ into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from OBJ: NETWORKS, NETWORKS_SUBNETS
        """
        LOG.debug("NETWORKS: %s", dict(obj))

        row_data = NeutronDriver.convert_objs(obj['networks'],
                                              self.networks_translator)
        network_tables = (self.NETWORKS, self.NETWORKS_SUBNETS)
        for table in network_tables:
            self.state[table] = set()
        for table, row in row_data:
            assert table in network_tables
            self.state[table].add(row)
        LOG.debug("NETWORKS: %s", self.state[self.NETWORKS])
        LOG.debug("NETWORKS_SUBNETS: %s", self.state[self.NETWORKS_SUBNETS])

    def _translate_ports(self, obj):
        """Translate the ports represented by OBJ into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from OBJ: PORTS, PORTS_ADDR_PAIRS,
        PORTS_SECURITY_GROUPS, PORTS_BINDING_CAPABILITIES,
        PORTS_FIXED_IPS, PORTS_FIXED_IPS_GROUPS,
        PORTS_EXTRA_DHCP_OPTS.
        """
        LOG.debug("PORTS: %s", obj)

        row_data = NeutronDriver.convert_objs(obj['ports'],
                                              self.ports_translator)
        port_tables = (self.PORTS, self.PORTS_ADDR_PAIRS,
                       self.PORTS_SECURITY_GROUPS,
                       self.PORTS_BINDING_CAPABILITIES, self.PORTS_FIXED_IPS,
                       self.PORTS_FIXED_IPS_GROUPS,
                       self.PORTS_EXTRA_DHCP_OPTS)
        for table in port_tables:
            self.state[table] = set()
        for table, row in row_data:
            assert table in port_tables
            self.state[table].add(row)
        for table in port_tables:
            LOG.debug('%s: %s', table, self.state[table])

    def _translate_routers(self, obj):
        """Translates the routers represented by OBJ into a single table.

        Assigns self.state[SECURITY_GROUPS] to that table.
        """
        LOG.debug("ROUTERS: %s", dict(obj))

        row_data = NeutronDriver.convert_objs(obj['routers'],
                                              self.routers_translator)
        router_tables = (self.ROUTERS, self.ROUTERS_EXTERNAL_GATEWAYS)
        for table in router_tables:
            self.state[table] = set()
        for table, row in row_data:
            assert table in router_tables
            self.state[table].add(row)
        for table in router_tables:
            LOG.debug('%s: %s', table, self.state[table])

    def _translate_security_groups(self, obj):
        LOG.debug("SECURITY_GROUPS: %s", dict(obj))

        row_data = NeutronDriver.convert_objs(obj['security_groups'],
                                              self.security_groups_translator)
        security_group_tables = (self.SECURITY_GROUPS,)
        for table in security_group_tables:
            self.state[table] = set()
        for table, row in row_data:
            assert table in security_group_tables
            self.state[table].add(row)
        for table in security_group_tables:
            LOG.debug('%s: %s', table, self.state[table])

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.neutron, action, action_args)

# Sample Mapping
# Network :
# ========
#
# json
# ------
# {u'status': u'ACTIVE', u'subnets':
#  [u'4cef03d0-1d02-40bb-8c99-2f442aac6ab0'],
# u'name':u'test-network', u'provider:physical_network': None,
# u'admin_state_up': True,
# u'tenant_id': u'570fe78a1dc54cffa053bd802984ede2',
# u'provider:network_type': u'gre',
# u'router:external': False, u'shared': False, u'id':
# u'240ff9df-df35-43ae-9df5-27fae87f2492',
# u'provider:segmentation_id': 4}
#
# tuple
# -----
#
# Networks : (u'ACTIVE', 'cdca5538-ae2d-11e3-92c1-bcee7bdf8d69',
# u'vova_network', None,
# True, u'570fe78a1dc54cffa053bd802984ede2', u'gre', 'False', 'False',
# u'1e3bc4fe-85c2-4b04-9b7f-ee40239787ef', 7)
#
# Networks and subnets
# ('cdcaa1a0-ae2d-11e3-92c1-bcee7bdf8d69',
#  u'4cef03d0-1d02-40bb-8c99-2f442aac6ab0')
#
#
# Ports
# ======
# json
# ----
# {u'status': u'ACTIVE',
# u'binding:host_id': u'havana', u'name': u'',
# u'allowed_address_pairs': [],
# u'admin_state_up': True, u'network_id':
# u'240ff9df-df35-43ae-9df5-27fae87f2492',
# u'tenant_id': u'570fe78a1dc54cffa053bd802984ede2',
# u'extra_dhcp_opts': [],
# u'binding:vif_type': u'ovs', u'device_owner':
# u'network:router_interface',
# u'binding:capabilities': {u'port_filter': True},
# u'mac_address': u'fa:16:3e:ab:90:df',
# u'fixed_ips': [{u'subnet_id':
# u'4cef03d0-1d02-40bb-8c99-2f442aac6ab0',
# u'ip_address': u'90.0.0.1'}], u'id':
# u'0a2ce569-85a8-45ec-abb3-0d4b34ff69ba',u'security_groups': [],
# u'device_id': u'864e4acf-bf8e-4664-8cf7-ad5daa95681e'},
# tuples
# -------
# Ports [(u'ACTIVE', u'havana', u'',
# '6425751e-ae2c-11e3-bba1-bcee7bdf8d69', 'True',
# u'240ff9df-df35-43ae-9df5-27fae87f2492',
# u'570fe78a1dc54cffa053bd802984ede2',
# '642579e2-ae2c-11e3-bba1-bcee7bdf8d69', u'ovs',
# u'network:router_interface', '64257dac-ae2c-11e3-bba1-bcee7bdf8d69',
# u'fa:16:3e:ab:90:df',
# '64258126-ae2c-11e3-bba1-bcee7bdf8d69',
# u'0a2ce569-85a8-45ec-abb3-0d4b34ff69ba',
# '64258496-ae2c-11e3-bba1-bcee7bdf8d69',
# u'864e4acf-bf8e-4664-8cf7-ad5daa95681e')
#
# Ports and Address Pairs
# [('6425751e-ae2c-11e3-bba1-bcee7bdf8d69', '')
# Ports and Security Groups
# [('64258496-ae2c-11e3-bba1-bcee7bdf8d69', '')
# Ports and Binding Capabilities [
# ('64257dac-ae2c-11e3-bba1-bcee7bdf8d69',u'port_filter','True')
# Ports and Fixed IPs [('64258126-ae2c-11e3-bba1-bcee7bdf8d69',
# u'subnet_id',u'4cef03d0-1d02-40bb-8c99-2f442aac6ab0'),
# ('64258126-ae2c-11e3-bba1-bcee7bdf8d69', u'ip_address',
# u'90.0.0.1')
#
# Ports and Extra dhcp opts [
# ('642579e2-ae2c-11e3-bba1-bcee7bdf8d69', '')



# Copyright 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging

from django.core.urlresolvers import reverse
from django.core.urlresolvers import reverse_lazy
from django.template.defaultfilters import dictsort
from django.utils.translation import ugettext_lazy as _
from horizon import exceptions
from horizon import forms
from horizon import messages
from horizon import tables
from openstack_dashboard.api import congress
import openstack_dashboard.dashboards.admin.datasources.utils as ds_utils
from openstack_dashboard.dashboards.admin.policies import (
    forms as policies_forms)
from openstack_dashboard.dashboards.admin.policies import (
    tables as policies_tables)
from openstack_dashboard.dashboards.admin.policies.rules import (
    tables as rules_tables)


LOG = logging.getLogger(__name__)


class IndexView(tables.DataTableView):
    """List policies."""
    table_class = policies_tables.PoliciesTable
    template_name = 'admin/policies/index.html'

    def get_data(self):
        try:
            policies = congress.policies_list(self.request)
        except Exception as e:
            msg = _('Unable to get policies list: %s') % e.message
            LOG.error(msg)
            messages.error(self.request, msg)
            return []
        return policies


class CreateView(forms.ModalFormView):
    form_class = policies_forms.CreatePolicy
    template_name = 'admin/policies/create.html'
    success_url = reverse_lazy('horizon:admin:policies:index')


class DetailView(tables.DataTableView):
    """List details about and rules in a policy."""
    table_class = rules_tables.PolicyRulesTable
    template_name = 'admin/policies/detail.html'

    def get_data(self):
        policy_name = self.kwargs['policy_name']
        try:
            policy_rules = congress.policy_rules_list(self.request,
                                                      policy_name)
        except Exception as e:
            msg_args = {'policy_name': policy_name, 'error': e.message}
            msg = _('Unable to get rules in policy "%(policy_name)s": '
                    '%(error)s') % msg_args
            LOG.error(msg)
            messages.error(self.request, msg)
            redirect = reverse('horizon:admin:policies:index')
            raise exceptions.Http302(redirect)

        for r in policy_rules:
            r.set_id_as_name_if_empty()
        return policy_rules

    def get_context_data(self, **kwargs):
        context = super(DetailView, self).get_context_data(**kwargs)
        policy_name = kwargs['policy_name']
        try:
            policy = congress.policy_get(self.request, policy_name)
        except Exception as e:
            msg_args = {'policy_name': policy_name, 'error': e.message}
            msg = _('Unable to get policy "%(policy_name)s": '
                    '%(error)s') % msg_args
            LOG.error(msg)
            messages.error(self.request, msg)
            redirect = reverse('horizon:admin:policies:index')
            raise exceptions.Http302(redirect)
        context['policy'] = policy

        # Alphabetize and convert list of data source tables and columns into
        # JSON formatted string consumable by JavaScript. Do this here instead
        # of in the Create Rule form so that the tables and columns lists
        # appear in the HTML document before the JavaScript that uses them.
        all_tables = ds_utils.get_datasource_tables(self.request)
        sorted_datasources = dictsort(all_tables, 'datasource')
        tables = []
        for ds in sorted_datasources:
            datasource_tables = ds['tables']
            datasource_tables.sort()
            for table in ds['tables']:
                tables.append('%s%s%s' % (ds['datasource'],
                                          congress.TABLE_SEPARATOR, table))
        context['tables'] = json.dumps(tables)

        datasource_columns = ds_utils.get_datasource_columns(self.request)
        sorted_datasources = dictsort(datasource_columns, 'datasource')
        columns = []
        for ds in sorted_datasources:
            sorted_tables = dictsort(ds['tables'], 'table')
            for tbl in sorted_tables:
                # Ignore service-derived tables, which are already included.
                if congress.TABLE_SEPARATOR in tbl['table']:
                    continue
                table_columns = tbl['columns']
                if table_columns:
                    table_columns.sort()
                else:
                    # Placeholder name for column when the table has none.
                    table_columns = ['_']

                for column in table_columns:
                    columns.append('%s%s%s %s' % (ds['datasource'],
                                                  congress.TABLE_SEPARATOR,
                                                  tbl['table'], column))
        context['columns'] = json.dumps(columns)
        return context



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging
import retrying

from congress.datalog import base
from congress.datalog import compile
from congress import harness
from congress.policy_engines import agnostic
from congress.tests import base as testbase
from congress.tests import helper

LOG = logging.getLogger(__name__)

NREC_THEORY = 'non-recursive theory'
DB_THEORY = 'database'
ACTION_THEORY = 'action'


class TestRuntimePerformance(testbase.TestCase):
    """Tests for Runtime performance that are not specific to any theory.

    To run one test:
      nosetests -v  \
      congress/tests/policy_engines/test_agnostic_performance.py:TestRuntimePerformance.test_foo

    To collect profiling data:
      python -m cProfile -o profile.out `which nosetests` -v \
      congress/tests/policy_engines/test_agnostic_performance.py:TestRuntimePerformance.test_foo

    To parse and sort profiling data in different ways:
      import pstats
      pstats.Stats('profile.out').strip_dirs().sort_stats("cum").print_stats()
      pstats.Stats('profile.out').strip_dirs().sort_stats("time").print_stats()
      pstats.Stats('profile.out').strip_dirs().sort_stats("calls").print_stats()

    """

    def setUp(self):
        super(TestRuntimePerformance, self).setUp()

        self._agnostic = agnostic.Runtime()
        self._agnostic.create_policy(NREC_THEORY,
                                     kind=base.NONRECURSIVE_POLICY_TYPE)
        self._agnostic.create_policy(DB_THEORY, kind=base.DATABASE_POLICY_TYPE)
        self._agnostic.debug_mode()
        self._agnostic.insert('', target=NREC_THEORY)

    def _create_event(self, table, tuple_, insert, target):
        return compile.Event(compile.Literal.create_from_table_tuple(table,
                                                                     tuple_),
                             insert=insert, target=target)

    def _create_large_tables(self, n, theory):
        facts = [compile.Fact('p', (i, j, k))
                 for i in range(n) for k in range(n) for j in range(n)]

        facts.extend(compile.Fact('q', (i,)) for i in range(n))
        self._agnostic.initialize_tables(['p', 'q'], facts, theory)

    def test_insert_nonrecursive(self):
        MAX = 100
        th = NREC_THEORY
        for i in range(MAX):
            self._agnostic.insert('r(%d)' % i, th)

    def test_insert_database(self):
        MAX = 100
        th = DB_THEORY
        for i in range(MAX):
            self._agnostic.insert('r(%d)' % i, th)

    def test_update_nonrecursive(self):
        MAX = 10000
        th = NREC_THEORY
        updates = [self._create_event('r', (i,), True, th)
                   for i in range(MAX)]
        self._agnostic.update(updates)

    def test_update_database(self):
        MAX = 1000
        th = DB_THEORY
        updates = [self._create_event('r', (i,), True, th)
                   for i in range(MAX)]
        self._agnostic.update(updates)

    def test_indexing(self):
        MAX = 100
        th = NREC_THEORY

        for table in ('a', 'b', 'c'):
            updates = [self._create_event(table, (i,), True, th)
                       for i in range(MAX)]
            self._agnostic.update(updates)

        # With indexing, this query should take O(n) time where n is MAX.
        # Without indexing, this query will take O(n^3).
        self._agnostic.insert('d(x) :- a(x), b(x), c(x)', th)
        ans = ' '.join(['d(%d)' % i for i in range(MAX)])
        self.assertTrue(helper.datalog_equal(self._agnostic.select('d(x)',
                                                                   th), ans))

    def test_runtime_initialize_tables(self):
        MAX = 700
        longstring = 'a' * 100
        facts = (compile.Fact('p',
                              (1, 2, 'foo', 'bar', i, longstring + str(i)))
                 for i in range(MAX))

        th = NREC_THEORY
        self._agnostic.initialize_tables(['p'], facts, th)

    def test_select_1match(self):
        # with different types of policies (exercise indexing, large sets,
        # many joins, etc)
        MAX = 10
        th = NREC_THEORY

        self._create_large_tables(MAX, th)
        self._agnostic.insert('r(x,y) :- p(x,x,y), q(x)', th)

        for i in range(100):
            # This select returns 1 result
            self._agnostic.select('r(1, 1)', th)

    def test_select_100matches(self):
        # with different types of policies (exercise indexing, large sets,
        # many joins, etc)
        MAX = 10
        th = NREC_THEORY

        self._create_large_tables(MAX, th)
        self._agnostic.insert('r(x,y) :- p(x,x,y), q(x)', th)

        # This takes about 60ms per select
        for i in range(10):
            # This select returns 100 results
            self._agnostic.select('r(x, y)', th)

    def test_simulate_latency(self):
        # We think the cost will be the sum of the simulate call + the cost to
        # do and undo the evaluation, so this test should focus on the cost
        # specific to the simulate call, so the the test should do a minimal
        # amount of evaluation.

        MAX = 10
        th = NREC_THEORY

        self._create_large_tables(MAX, th)
        self._agnostic.create_policy(ACTION_THEORY,
                                     kind=base.ACTION_POLICY_TYPE)

        self._agnostic.insert('q(0)', th)
        self._agnostic.insert('s(x) :- q(x), p(x,0,0)', th)

        # This takes about 13ms per simulate.  The query for s(x) can use
        # indexing, so it should be efficient.
        for _ in range(100):
            self._agnostic.simulate('s(x)', th, 'p-(0,0,0)',
                                    ACTION_THEORY, delta=True)

    def test_simulate_throughput(self):
        # up to 250 requests per second
        pass

    def test_update_rate(self):
        pass

    def test_concurrency(self):
        pass


class TestDsePerformance(testbase.SqlTestCase):

    def setUp(self):
        super(TestDsePerformance, self).setUp()
        self.cage = harness.create(helper.root_path(), config_override={})
        self.api = {'policy': self.cage.service_object('api-policy'),
                    'rule': self.cage.service_object('api-rule'),
                    'table': self.cage.service_object('api-table'),
                    'row': self.cage.service_object('api-row'),
                    'datasource': self.cage.service_object('api-datasource'),
                    'status': self.cage.service_object('api-status'),
                    'schema': self.cage.service_object('api-schema')}
        self.engine = self.cage.service_object('engine')

    @retrying.retry(wait_fixed=100)
    def wait_til_query_nonempty(self, query, policy):
        if len(self.engine.select(query, target=policy)) == 0:
            raise Exception("Query %s is not empty" % query)

    def test_initialize_tables_dse(self):
        """Test performance of initializing data with DSE and Engine.

        This test populates the tables exported by a datasource driver,
        and then invokes the poll() method to send that data to the
        policy engine.  It tests the amount of time to send tables
        across the DSE and load them into the policy engine.
        """
        MAX_TUPLES = 700
        # install datasource driver we can control
        self.cage.loadModule(
            "TestDriver",
            helper.data_module_path(
                "../tests/datasources/test_driver.py"))
        self.cage.createservice(
            name="data",
            moduleName="TestDriver",
            args=helper.datasource_openstack_args())
        driver = self.cage.service_object('data')
        driver.poll_time = 0
        self.engine.create_policy('data')

        # set return value for datasource driver
        facts = [(1, 2.3, 'foo', 'bar', i, 'a'*100 + str(i))
                 for i in range(MAX_TUPLES)]
        driver.state = {'p': facts}

        # Send formula to engine (so engine subscribes to data:p)
        policy = self.engine.DEFAULT_THEORY
        formula = compile.parse1(
            'q(1) :- data:p(1, 2.3, "foo", "bar", 1, %s)' % ('a'*100 + '1'))
        self.api['rule'].publish(
            'policy-update', [compile.Event(formula, target=policy)])

        # Poll data and wait til it arrives at engine
        driver.poll()
        self.wait_til_query_nonempty('q(1)', policy)

    def test_initialize_tables_full(self):
        """Test performance of initializing data with Datasource, DSE, Engine.

        This test gives a datasource driver the Python data that would
        have resulted from making an API call and parsing it into Python
        and then polls that datasource, waiting until the data arrives
        in the policy engine.  It tests the amount of time required to
        translate Python data into tables, send those tables over the DSE,
        and load them into the policy engine.
        """
        MAX_TUPLES = 700
        # install datasource driver we can control
        self.cage.loadModule(
            "PerformanceTestDriver",
            helper.data_module_path(
                "../tests/datasources/performance_datasource_driver.py"))
        self.cage.createservice(
            name="data",
            moduleName="PerformanceTestDriver",
            args=helper.datasource_openstack_args())
        driver = self.cage.service_object('data')
        driver.poll_time = 0
        self.engine.create_policy('data')

        # set return value for datasource driver
        facts = [{'field1': 1,
                  'field2': 2.3,
                  'field3': 'foo',
                  'field4': 'bar',
                  'field5': i,
                  'field6': 'a'*100 + str(i)}
                 for i in range(MAX_TUPLES)]
        driver.client_data = facts

        # Send formula to engine (so engine subscribes to data:p)
        policy = self.engine.DEFAULT_THEORY
        formula = compile.parse1(
            'q(1) :- data:p(1, 2.3, "foo", "bar", 1, %s)' % ('a'*100 + '1'))
        LOG.info("publishing rule")
        self.api['rule'].publish(
            'policy-update', [compile.Event(formula, target=policy)])

        # Poll data and wait til it arrives at engine
        driver.poll()
        self.wait_til_query_nonempty('q(1)', policy)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.datasources import plexxi_driver
from congress.tests import base
from congress.tests.datasources import plexxi_fakes
from congress.tests import helper


class TestPlexxiDriver(base.TestCase):
    def setUp(self):
        super(TestPlexxiDriver, self).setUp()
        args = helper.datasource_openstack_args()
        args['unique_names'] = 'False'
        session = plexxi_fakes.MockCoreSession()
        self.driver = plexxi_driver.PlexxiDriver(args=args, session=session)
        self.driver.exchange = True
        vnic1 = plexxi_fakes.MockNIC(
            uuid='f318ac0a-9255-4af0-8a41-6f3fbc06c8aa',
            mac='B8:ED:0A:4D:82:91')
        vnic2 = plexxi_fakes.MockNIC(
            uuid='f318ac0a-9255-4af0-8a41-6f3fbc06c8a2',
            mac='B8:ED:0A:4D:82:99')
        pnic1 = plexxi_fakes.MockNIC(
            uuid='f318ac0a-9255-4af0-8a41-6f3fbc06c8ab',
            mac='B8:ED:0A:4D:82:92')
        pnic2 = plexxi_fakes.MockNIC(
            uuid='f318ac0a-9255-4af0-8a41-6f3fbc06c8ac',
            mac='B8:ED:0A:4D:82:93')
        host1 = plexxi_fakes.MockHost('eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                                      'mock1',
                                      1,
                                      [pnic1])
        vm1 = plexxi_fakes.MockVM('2ca924f6-90aa-4ce8-a986-f62f8f64d14b',
                                  '192.168.90.2',
                                  'namevm',
                                  host1,
                                  [vnic1])
        host1.addvm(vm1)
        switch1 = plexxi_fakes.MockSwitch(
            '12da13e3-ecb2-4c26-98a0-26cb07f9c33d',
            '192.168.90.3',
            'switch1',
            'HEALTHY',
            [pnic2])
        affinity = plexxi_fakes.MockAffinity(
            'fd487ecf-5279-4d3c-9378-7fb214f5dd5a', 'Testfinnity')
        affinity2 = plexxi_fakes.MockAffinity(
            'fd487ecf-5279-4d3c-9378-7fb214f5dd5b', 'Testfinnity2')
        vswitch = plexxi_fakes.MockVSwitch(
            'fd487ecf-5279-4d3c-9378-7fb214f5dd5c',
            [host1],
            [vnic2])
        link1 = plexxi_fakes.MockNetworkLink(
            'fd487ecf-5279-4d3c-9378-7fb214f5dd5f',
            'Link1',
            host1,
            switch1)
        port = plexxi_fakes.MockPort('fd487ecf-5279-4d3c-9378-7fb214f5dd5d',
                                     'Port1',
                                     [link1])
        port2 = plexxi_fakes.MockPort('fd487ecf-5279-4d3c-9378-7fb214f5dd5e',
                                      'Port2',
                                      None)

        self.hosts = [host1]
        self.pswitches = [switch1]
        self.affinites = [affinity, affinity2]
        self.vswitches = [vswitch]
        self.vms = [vm1]
        self.ports = [port, port2]

    def test_translate_hosts(self):
        self.driver._translate_hosts(self.hosts)
        ExpectedHosts = [('eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                          'mock1',
                          1,
                          1)]
        self.assertEqual(ExpectedHosts, self.driver.hosts)
        ExpectedHost_Macs = [('eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                             'B8:ED:0A:4D:82:92')]
        self.assertEqual(ExpectedHost_Macs, self.driver.mac_list)
        ExpectedGuests = [('eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                          '2ca924f6-90aa-4ce8-a986-f62f8f64d14b')]
        self.assertEqual(ExpectedGuests, self.driver.guest_list)

    def test_translate_pswitches(self):
        self.driver._translate_pswitches(self.pswitches)
        ExpectedpSwitches = [('12da13e3-ecb2-4c26-98a0-26cb07f9c33d',
                              '192.168.90.3',
                              'HEALTHY')]
        self.assertEqual(self.driver.plexxi_switches, ExpectedpSwitches)
        ExpectedPSmacs = [('12da13e3-ecb2-4c26-98a0-26cb07f9c33d',
                           'B8:ED:0A:4D:82:93')]
        self.assertEqual(ExpectedPSmacs, self.driver.ps_macs)

    def test_translate_affinites(self):
        self.driver._translate_affinites(self.affinites)
        ExpectedAffinities = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5a',
                              'Testfinnity'),
                              ('fd487ecf-5279-4d3c-9378-7fb214f5dd5b',
                              'Testfinnity2')]

        self.assertEqual(ExpectedAffinities, self.driver.affinities)

    def test_translate_vswitches(self):
        self.driver._translate_vswitches(self.vswitches)
        ExpectedvSwitches = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5c',
                              1,
                              1)]
        self.assertEqual(ExpectedvSwitches, self.driver.vswitches)
        ExpectedvSwitch_macs = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5c',
                                 'B8:ED:0A:4D:82:99')]
        self.assertEqual(ExpectedvSwitch_macs, self.driver.vswitch_macs)
        ExpectedvSwitch_hosts = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5c',
                                  'eed4ebfc-25e5-4a65-9f37-b70b8e8219d3')]

        self.assertEqual(ExpectedvSwitch_hosts, self.driver.vswitch_hosts)

    def test_translate_vms(self):
        self.driver._translate_vms(self.vms)
        ExpectedVMs = [('2ca924f6-90aa-4ce8-a986-f62f8f64d14b',
                        'namevm',
                        'eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                        '192.168.90.2',
                        1)]

        self.assertEqual(ExpectedVMs, self.driver.vms)
        Expectedvm_macs = [('2ca924f6-90aa-4ce8-a986-f62f8f64d14b',
                            'B8:ED:0A:4D:82:91')]
        self.assertEqual(Expectedvm_macs, self.driver.vm_macs)

    def test_translate_ports(self):
        self.driver._translate_ports(self.ports)
        ExpectedPorts = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5d',
                          'Port1'),
                         ('fd487ecf-5279-4d3c-9378-7fb214f5dd5e',
                          'Port2')]
        self.assertEqual(ExpectedPorts, self.driver.ports)
        ExpectedLinks = [('fd487ecf-5279-4d3c-9378-7fb214f5dd5f',
                          'Link1',
                          'fd487ecf-5279-4d3c-9378-7fb214f5dd5d',
                          '12da13e3-ecb2-4c26-98a0-26cb07f9c33d',
                          'switch1',
                          'eed4ebfc-25e5-4a65-9f37-b70b8e8219d3',
                          'mock1')]
        self.assertEqual(ExpectedLinks, self.driver.network_links)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

try:
    # For Python 3
    import http.client as httplib
except ImportError:
    import httplib

# TODO(thinrichs): move this out of api directory.  Could go into
#    the exceptions.py file.  The HTTP error codes may make these errors
#    look like they are only useful for the API, but actually they are
#    just encoding the classification of the error using http codes.
#    To make this more explicit, we could have 2 dictionaries where
#    one maps an error name (readable for programmers) to an error number
#    and another dictionary that maps an error name/number to the HTTP
#    classification.  But then it would be easy for a programmer when
#    adding a new error to forget one or the other.

# name of unknown error
UNKNOWN = 'unknown'


# dict mapping error name to (<error id>, <description>, <http error code>)
errors = {}
errors[UNKNOWN] = (
    1000, "Unknown error", httplib.BAD_REQUEST)
errors['add_item_id'] = (
    1001, "Add item does not support user-chosen ID", httplib.BAD_REQUEST)
errors['rule_syntax'] = (
    1002, "Syntax error for rule", httplib.BAD_REQUEST)
errors['multiple_rules'] = (
    1003, "Received string representing more than 1 rule", httplib.BAD_REQUEST)
errors['incomplete_simulate_args'] = (
    1004, "Simulate requires parameters: query, sequence, action_policy",
    httplib.BAD_REQUEST)
errors['simulate_without_policy'] = (
    1005, "Simulate must be told which policy evaluate the query on",
    httplib.BAD_REQUEST)
errors['sequence_syntax'] = (
    1006, "Syntax error in sequence", httplib.BAD_REQUEST)
errors['simulate_error'] = (
    1007, "Error in simulate procedure", httplib.INTERNAL_SERVER_ERROR)
errors['rule_already_exists'] = (
    1008, "Rule already exists", httplib.CONFLICT)
errors['schema_get_item_id'] = (
    1009, "Get item for schema does not support user-chosen ID",
    httplib.BAD_REQUEST)
errors['policy_name_must_be_provided'] = (
    1010, "A name must be provided when creating a policy",
    httplib.BAD_REQUEST)
errors['no_policy_update_owner'] = (
    1012, "The policy owner_id cannot be updated",
    httplib.BAD_REQUEST)
errors['no_policy_update_kind'] = (
    1013, "The policy kind cannot be updated",
    httplib.BAD_REQUEST)
errors['failed_to_create_policy'] = (
    1014, "A new policy could not be created",
    httplib.INTERNAL_SERVER_ERROR)
errors['policy_id_must_not_be_provided'] = (
    1015, "An ID may not be provided when creating a policy",
    httplib.BAD_REQUEST)
errors['execute_error'] = (
    1016, "Error in execution procedure", httplib.INTERNAL_SERVER_ERROR)
errors['service_action_syntax'] = (
    1017, "Incorrect action syntax. Requires: <service>:<action>",
    httplib.BAD_REQUEST)
errors['execute_action_args_syntax'] = (
    1018, "Incorrect argument syntax. "
    "Requires: {'positional': [<args>], 'named': {<key>:<value>,}}",
    httplib.BAD_REQUEST)
errors['rule_not_permitted'] = (
    1019, "Rules not permitted on non persisted policies.",
    httplib.BAD_REQUEST)
errors['policy_not_exist'] = (
    1020, "The specified policy does not exist.", httplib.BAD_REQUEST)
errors['policy_rule_insertion_failure'] = (
    1021, "The policy rule could not be inserted.", httplib.BAD_REQUEST)
errors['policy_abbreviation_error'] = (
    1022, "The policy abbreviation must be a string and the length of the "
          "string must be equal to or less than 5 characters.",
    httplib.BAD_REQUEST)


def get(name):
    if name not in errors:
        name = UNKNOWN
    return errors[name][:2]


def get_num(name):
    if name not in errors:
        name = UNKNOWN
    return errors[name][0]


def get_desc(name):
    if name not in errors:
        name = UNKNOWN
    return errors[name][1]


def get_http(name):
    if name not in errors:
        name = UNKNOWN
    return errors[name][2]



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import ironicclient.v1.chassis as IrChassis
import ironicclient.v1.driver as IrDriver
import ironicclient.v1.node as IrNode
import ironicclient.v1.port as IrPort
import mock

from congress.datasources import ironic_driver
from congress.tests import base
from congress.tests import helper


class TestIronicDriver(base.TestCase):

    def setUp(self):
        super(TestIronicDriver, self).setUp()
        self.ironic_client_p = mock.patch("ironicclient.client.get_client")
        self.ironic_client_p.start()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['client'] = mock.MagicMock()

        self.driver = ironic_driver.IronicDriver(args=args)

        self.mock_chassis = {"chassis": [
            {"uuid": "89a15e07-5c80-48a4-b440-9c61ddb7e652",
             "extra": {},
             "created_at": "2015-01-13T06:52:01+00:00",
             "updated_at": None,
             "description": "ironic test chassis"}]}

        self.mock_nodes = {"nodes": [
            {"instance_uuid": "2520745f-b4da-4e10-9d32-84451cfa8b33",
             "uuid": "9cf035f0-351c-43d5-8968-f9fe2c41787b",
             "chassis_uuid": "89a15e07-5c80-48a4-b440-9c61ddb7e652",
             "properties": {"memory_mb": "512", "cpu_arch": "x86_64",
                            "local_gb": "10", "cpus": "1"},
             "driver": "pxe_ssh",
             "maintenance": False,
             "console_enabled": False,
             "created_at": "2015-01-13T06:52:02+00:00",
             "updated_at": "2015-02-10T07:55:23+00:00",
             "provision_updated_at": "2015-01-13T07:55:24+00:00",
             "provision_state": "active",
             "power_state": "power on"},
            {"instance_uuid": None,
             "uuid": "7a95ebf5-f213-4427-b669-010438f43e87",
             "chassis_uuid": "89a15e07-5c80-48a4-b440-9c61ddb7e652",
             "properties": {"memory_mb": "512", "cpu_arch": "x86_64",
                            "local_gb": "10", "cpus": "1"},
             "driver": "pxe_ssh",
             "maintenance": False,
             "console_enabled": False,
             "created_at": "2015-01-13T06:52:04+00:00",
             "updated_at": "2015-02-10T07:55:24+00:00",
             "provision_updated_at": None,
             "provision_state": None,
             "power_state": "power off"}]}

        self.mock_ports = {"ports": [
            {"uuid": "43190aae-d5fe-444f-9d50-155fca4bad82",
             "node_uuid": "9cf035f0-351c-43d5-8968-f9fe2c41787b",
             "extra": {"vif_port_id": "9175f72b-5783-4cea-8ae0-55df69fee568"},
             "created_at": "2015-01-13T06:52:03+00:00",
             "updated_at": "2015-01-30T03:17:23+00:00",
             "address": "52:54:00:7f:e7:2e"},
            {"uuid": "49f3205a-db1e-4497-9371-6011ef572981",
             "node_uuid": "7a95ebf5-f213-4427-b669-010438f43e87",
             "extra": {},
             "created_at": "2015-01-13T06:52:05+00:00",
             "updated_at": None,
             "address": "52:54:00:98:f2:4e"}]}

        self.mock_drivers = {"drivers": [
            {"hosts": ["localhost"], "name": "pxe_ssh"},
            {"hosts": ["localhost"], "name": "pxe_ipmitool"},
            {"hosts": ["localhost"], "name": "fake"}]}

        self.expected_state = {
            'drivers': set([
                ('pxe_ipmitool',),
                ('fake',),
                ('pxe_ssh',)]),
            'node_properties': set([
                ('7a95ebf5-f213-4427-b669-010438f43e87',
                 '512', 'x86_64', '10', '1'),
                ('9cf035f0-351c-43d5-8968-f9fe2c41787b',
                 '512', 'x86_64', '10', '1')]),
            'chassises': set([
                ('89a15e07-5c80-48a4-b440-9c61ddb7e652',
                 '2015-01-13T06:52:01+00:00', 'None')]),
            'active_hosts': set([
                ('pxe_ipmitool', 'localhost'),
                ('pxe_ssh', 'localhost'),
                ('fake', 'localhost')]),
            'nodes': set([
                ('9cf035f0-351c-43d5-8968-f9fe2c41787b',
                 '89a15e07-5c80-48a4-b440-9c61ddb7e652',
                 'power on',
                 'False',
                 'pxe_ssh',
                 '2520745f-b4da-4e10-9d32-84451cfa8b33',
                 '2015-01-13T06:52:02+00:00',
                 '2015-01-13T07:55:24+00:00',
                 '2015-02-10T07:55:23+00:00'),
                ('7a95ebf5-f213-4427-b669-010438f43e87',
                 '89a15e07-5c80-48a4-b440-9c61ddb7e652',
                 'power off',
                 'False',
                 'pxe_ssh',
                 'None',
                 '2015-01-13T06:52:04+00:00',
                 'None',
                 '2015-02-10T07:55:24+00:00')]),
            'ports': set([
                ('49f3205a-db1e-4497-9371-6011ef572981',
                 '7a95ebf5-f213-4427-b669-010438f43e87',
                 '52:54:00:98:f2:4e', '',
                 '2015-01-13T06:52:05+00:00', 'None'),
                ('43190aae-d5fe-444f-9d50-155fca4bad82',
                 '9cf035f0-351c-43d5-8968-f9fe2c41787b',
                 '52:54:00:7f:e7:2e',
                 '9175f72b-5783-4cea-8ae0-55df69fee568',
                 '2015-01-13T06:52:03+00:00',
                 '2015-01-30T03:17:23+00:00')])
        }

    def mock_value(self, mock_data, key, obj_class):
        data = mock_data[key]
        return [obj_class(self, res, loaded=True) for res in data if res]

    def test_driver_called(self):
        self.assertIsNotNone(self.driver.ironic_client)

    def test_update_from_datasource(self):
        with base.nested(
            mock.patch.object(self.driver.ironic_client.chassis,
                              "list",
                              return_value=self.mock_value(self.mock_chassis,
                                                           "chassis",
                                                           IrChassis.Chassis)),
            mock.patch.object(self.driver.ironic_client.node,
                              "list",
                              return_value=self.mock_value(self.mock_nodes,
                                                           "nodes",
                                                           IrNode.Node)),
            mock.patch.object(self.driver.ironic_client.port,
                              "list",
                              return_value=self.mock_value(self.mock_ports,
                                                           "ports",
                                                           IrPort.Port)),
            mock.patch.object(self.driver.ironic_client.driver,
                              "list",
                              return_value=self.mock_value(self.mock_drivers,
                                                           "drivers",
                                                           IrDriver.Driver)),
            ) as (self.driver.ironic_client.chassis.list,
                  self.driver.ironic_client.node.list,
                  self.driver.ironic_client.port.list,
                  self.driver.ironic_client.driver.list):
            self.driver.update_from_datasource()
            self.assertEqual(self.expected_state, self.driver.state)

    def test_execute(self):
        class IronicClient(object):
            def __init__(self):
                self.testkey = None

            def updateNode(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        ironic_client = IronicClient()
        self.driver.ironic_client = ironic_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('updateNode', api_args)

        self.assertEqual(expected_ans, ironic_client.testkey)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_policy

Revision ID: 532e9e1f0f3a
Revises: initial_db
Create Date: 2014-12-18 14:52:20.402861

"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# revision identifiers, used by Alembic.
revision = '532e9e1f0f3a'
down_revision = 'initial_db'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('policies',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('created_at', sa.DateTime(), nullable=False),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted', sa.String(length=36),
                              server_default='', nullable=True),
                    sa.Column('name', sa.Text(), nullable=False),
                    sa.Column('abbreviation', sa.String(length=5),
                              nullable=False),
                    sa.Column('description', sa.Text(), nullable=False),
                    sa.Column('owner', sa.Text(), nullable=False),
                    sa.Column('kind', sa.Text(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')


def downgrade():
    op.drop_table('policies')



# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from ceilometerclient.v2 import client
import mock

from congress.datasources import ceilometer_driver
from congress.tests import base
from congress.tests.datasources import util
from congress.tests import helper

ResponseObj = util.ResponseObj


class TestCeilometerDriver(base.TestCase):

    def setUp(self):
        super(TestCeilometerDriver, self).setUp()
        self.ceilometer_client = mock.MagicMock()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        with mock.patch.object(client.Client, '__init__',
                               return_value=None):
            self.driver = ceilometer_driver.CeilometerDriver(
                name='testceilometer',
                args=args)

    def test_list_meters(self):
        meters_data = [
            ResponseObj({'name': 'instance:m1.tiny',
                         'type': 'gauge',
                         'unit': 'instance',
                         'resource_id': 'a257ba13-0b36-4a86-ae89-f78dd28b8ae5',
                         'user_id': '2b01323fd71345bc8cdc5dbbd6d127ea',
                         'project_id': '0020df0171ec41b597cd8b3002e21bee',
                         'meter_id': 'YTI1N2JhMTMtMGIzNi00YTg2LWFlODktZjc4ZG',
                         'source': 'openstack'}),
            ResponseObj({'name': 'network.incoming.bytes',
                         'type': 'cumulative',
                         'unit': 'B',
                         'resource_id': 'instance-00000001-tap437ce69c-e5',
                         'user_id': '2b01323fd71345bc8cdc5dbbd6d127ea',
                         'project_id': '0020df0171ec41b597cd8b3002e21bee',
                         'meter_id': 'aW5zdGFuY2UtMDAwMDAwMDEtYTI1N2JhMT',
                         'source': 'openstack'})]

        self.driver._translate_meters(meters_data)
        meter_list = list(self.driver.state['meters'])
        self.assertIsNotNone(meter_list)
        self.assertEqual(2, len(meter_list))

        for meter in meter_list:
            if meter[1] == 'network.incoming.bytes':
                meter1 = meter
            elif meter[1] == 'instance:m1.tiny':
                meter2 = meter

        # Verifying individual tuple data
        self.assertEqual(('aW5zdGFuY2UtMDAwMDAwMDEtYTI1N2JhMT',
                          'network.incoming.bytes',
                          'cumulative',
                          'B',
                          'openstack',
                          'instance-00000001-tap437ce69c-e5',
                          '2b01323fd71345bc8cdc5dbbd6d127ea',
                          '0020df0171ec41b597cd8b3002e21bee'),
                         meter1)

        self.assertEqual(('YTI1N2JhMTMtMGIzNi00YTg2LWFlODktZjc4ZG',
                          'instance:m1.tiny',
                          'gauge',
                          'instance',
                          'openstack',
                          'a257ba13-0b36-4a86-ae89-f78dd28b8ae5',
                          '2b01323fd71345bc8cdc5dbbd6d127ea',
                          '0020df0171ec41b597cd8b3002e21bee'),
                         meter2)

    def test_list_alarms(self):
        threshold_rule1 = {'key1': 'value1',
                           'key2': 'value2',
                           'key3': 'value3'}
        threshold_rule2 = {'key4': 'value4',
                           'key5': 'value5',
                           'key6': 'value6'}

        alarms_data = [
            ResponseObj({'alarm_id': '7ef99553-a73f-4b18-a617-997a479c48e9',
                         'name': 'cpu_high2',
                         'state': 'insufficient data',
                         'enabled': 'True',
                         'threshold_rule': threshold_rule1,
                         'type': 'threshold',
                         'description': 'instance running hot',
                         'time_constraints': '[]',
                         'user_id': '2b01323fd71345bc8cdc5dbbd6d127ea',
                         'project_id': '',
                         'alarm_actions': "[u'log://']",
                         'ok_actions': '[]',
                         'insufficient_data_actions': '[]',
                         'repeat_actions': 'False',
                         'timestamp': '2014-09-30T05:00:43.351041',
                         'state_timestamp': '2014-09-30T05:00:43.351041'}),
            ResponseObj({'alarm_id': 'd1b2b7a7-9512-4290-97ca-2580ed72c375',
                         'name': 'cpu_high',
                         'state': 'insufficient data',
                         'enabled': 'True',
                         'threshold_rule': threshold_rule2,
                         'type': 'threshold',
                         'description': 'instance running hot',
                         'time_constraints': '[]',
                         'user_id': '2b01323fd71345bc8cdc5dbbd6d127ea',
                         'project_id': '',
                         'alarm_actions': "[u'log://']",
                         'ok_actions': '[]',
                         'insufficient_data_actions': '[]',
                         'repeat_actions': 'False',
                         'timestamp': '2014-09-30T04:55:36.015925',
                         'state_timestamp': '2014-09-30T04:55:36.015925'})]

        self.driver._translate_alarms(alarms_data)
        alarm_list = list(self.driver.state['alarms'])
        self.assertIsNotNone(alarm_list)
        self.assertEqual(2, len(alarm_list))

        alarm_threshold_rule = list(self.driver.state['alarms.threshold_rule'])
        self.assertIsNotNone(alarm_threshold_rule)
        self.assertEqual(6, len(alarm_threshold_rule))

        for alarm in alarm_list:
            if alarm[1] == 'cpu_high2':
                alarm1 = alarm
            elif alarm[1] == 'cpu_high':
                alarm2 = alarm

        for thres in alarm_threshold_rule:
            if thres[1] in ['key1', 'key2', 'key3']:
                thresh_rule_id1 = thres[0]
            elif thres[1] in ['key4', 'key5', 'key6']:
                thresh_rule_id2 = thres[0]

        # Verifying individual tuple data
        self.assertEqual(('7ef99553-a73f-4b18-a617-997a479c48e9',
                          'cpu_high2', 'insufficient data',
                          'True',
                          thresh_rule_id1,
                          'threshold', 'instance running hot',
                          '[]', '2b01323fd71345bc8cdc5dbbd6d127ea',
                          '', "[u'log://']", '[]', '[]', 'False',
                          '2014-09-30T05:00:43.351041',
                          '2014-09-30T05:00:43.351041'),
                         alarm1)
        self.assertEqual(('d1b2b7a7-9512-4290-97ca-2580ed72c375',
                          'cpu_high', 'insufficient data', 'True',
                          thresh_rule_id2,
                          'threshold', 'instance running hot',
                          '[]', '2b01323fd71345bc8cdc5dbbd6d127ea',
                          '', "[u'log://']", '[]', '[]',
                          'False', '2014-09-30T04:55:36.015925',
                          '2014-09-30T04:55:36.015925'),
                         alarm2)

    def test_list_events(self):
        trait1 = [{'name': 'value1', 'type': 'value2', 'value': 'value3'},
                  {'name': 'value7', 'type': 'value8', 'value': 'value9'}]
        trait2 = [{'name': 'value4', 'type': 'value5', 'value': 'value6'}]
        trait3 = []

        events_data = [
            ResponseObj({'message_id': '6834861c-ccb3-4c6f-ac00-fe8fe1ad4ed4',
                         'event_type': 'image.create',
                         'generated': '2014-09-29T08:19:45.556301',
                         'traits': trait1}),
            ResponseObj({'message_id': '3676d6d4-5c65-4442-9eda-b78d750ea91f',
                         'event_type': 'compute.instance.update',
                         'generated': '2014-09-30T04:54:45.395522',
                         'traits': trait2}),
            ResponseObj({'message_id': 'fae7b03d-b5b7-4b4f-b2ef-06d2af03f21e',
                         'event_type': 'telemetry.api',
                         'generated': '2015-09-02T10:12:50.338919',
                         'traits': trait3})]

        self.driver._translate_events(events_data)
        events_set = self.driver.state['events']
        expected_events = {('6834861c-ccb3-4c6f-ac00-fe8fe1ad4ed4',
                            'image.create', '2014-09-29T08:19:45.556301'),
                           ('3676d6d4-5c65-4442-9eda-b78d750ea91f',
                            'compute.instance.update',
                            '2014-09-30T04:54:45.395522'),
                           ('fae7b03d-b5b7-4b4f-b2ef-06d2af03f21e',
                            'telemetry.api', '2015-09-02T10:12:50.338919')}
        self.assertEqual(expected_events, events_set)

        event_traits_set = self.driver.state['events.traits']
        expected_traits = {('6834861c-ccb3-4c6f-ac00-fe8fe1ad4ed4',
                            'value1', 'value2', 'value3'),
                           ('6834861c-ccb3-4c6f-ac00-fe8fe1ad4ed4',
                            'value7', 'value8', 'value9'),
                           ('3676d6d4-5c65-4442-9eda-b78d750ea91f',
                            'value4', 'value5', 'value6')}
        self.assertEqual(expected_traits, event_traits_set)

    def test_list_statistics(self):
        statistics_data = [
            {'meter_name': 'network',
             'period': 0, 'groupby':
             {'resource_id': '2fdef98a-8a00-4094-b6b8-b3f742076417'},
             'period_start': '2014-12-09T12:52:39.366015',
             'period_end': '2014-12-09T12:52:56.478338',
             'max': 0.0, 'min': 0.0, 'avg': 0.0, 'sum': 0.0,
             'count': 10, 'duration': 17.112323, 'unit': 'GB',
             'duration_start': '2014-12-09T12:52:39.366015',
             'duration_end': '2014-12-09T12:52:56.478338'},
            {'meter_name': 'instance',
             'period': 0, 'groupby':
             {'resource_id': '8a1340fa-fd43-4376-9deb-37c872c47e38'},
             'period_start': '2014-12-09T12:52:39.366015',
             'period_end': '2014-12-09T13:04:34',
             'max': 1.0, 'min': 1.0, 'avg': 1.0, 'sum': 13.0,
             'count': 13, 'duration': 714.633985,
             'unit': 'instance',
             'duration_start': '2014-12-09T12:52:39.366015',
             'duration_end': '2014-12-09T13:04:34'}]

        self.driver._translate_statistics(statistics_data)
        statistics_list = list(self.driver.state['statistics'])
        self.assertIsNotNone(statistics_list)
        self.assertEqual(2, len(statistics_list))

        # Verifying individual tuple data
        s1 = next(x for x in statistics_list if x[0] == 'network')
        s2 = next(x for x in statistics_list if x[0] == 'instance')

        self.assertEqual(('network',
                          '2fdef98a-8a00-4094-b6b8-b3f742076417',
                          0.0, 10, 17.112323,
                          '2014-12-09T12:52:39.366015',
                          '2014-12-09T12:52:56.478338',
                          0.0, 0.0, 0,
                          '2014-12-09T12:52:56.478338',
                          '2014-12-09T12:52:39.366015',
                          0.0, 'GB'), s1)
        self.assertEqual(('instance',
                          '8a1340fa-fd43-4376-9deb-37c872c47e38',
                          1.0, 13, 714.633985,
                          '2014-12-09T12:52:39.366015',
                          '2014-12-09T13:04:34',
                          1.0, 1.0, 0,
                          '2014-12-09T13:04:34',
                          '2014-12-09T12:52:39.366015',
                          13.0, 'instance'), s2)

    def test_execute(self):
        class CeilometerClient(object):
            def __init__(self):
                self.testkey = None

            def setAlarm(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        ceilometer_client = CeilometerClient()
        self.driver.ceilometer_client = ceilometer_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('setAlarm', api_args)

        self.assertEqual(ceilometer_client.testkey, expected_ans)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from tempest import clients
from tempest import config
from tempest.lib import exceptions
from tempest import test

from congress_tempest_tests.tests.scenario import manager_congress


CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestCeilometerDriver(manager_congress.ScenarioPolicyBase):

    @classmethod
    def skip_checks(cls):
        super(TestCeilometerDriver, cls).skip_checks()
        if not CONF.service_available.ceilometer:
            msg = ("%s skipped as ceilometer is not available" %
                   cls.__class__.__name__)
            raise cls.skipException(msg)

    def setUp(cls):
        super(TestCeilometerDriver, cls).setUp()
        cls.os = clients.Manager(cls.admin_manager.auth_provider.credentials)
        cls.telemetry_client = cls.os.telemetry_client
        cls.datasource_id = manager_congress.get_datasource_id(
            cls.admin_manager.congress_client, 'ceilometer')

    @test.attr(type='smoke')
    def test_ceilometer_meters_table(self):
        meter_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'meters')['columns'])
        meter_id_col = next(i for i, c in enumerate(meter_schema)
                            if c['name'] == 'meter_id')

        def _check_data_table_ceilometer_meters():
            # Fetch data from ceilometer each time, because this test may start
            # before ceilometer has all the users.
            meters = self.telemetry_client.list_meters()
            meter_map = {}
            for meter in meters:
                meter_map[meter['meter_id']] = meter

            results = (
                self.admin_manager.congress_client.list_datasource_rows(
                    self.datasource_id, 'meters'))
            for row in results['results']:
                try:
                    meter_row = meter_map[row['data'][meter_id_col]]
                except KeyError:
                    return False
                for index in range(len(meter_schema)):
                    if (str(row['data'][index]) !=
                            str(meter_row[meter_schema[index]['name']])):
                        return False
            return True

        if not test.call_until_true(func=_check_data_table_ceilometer_meters,
                                    duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
from oslo_config import cfg
from oslo_utils import uuidutils

from congress.api import error_codes
from congress.api import policy_model
from congress.api import webservice
from congress.dse2 import dse_node
from congress import harness
from congress.tests import base
from congress.tests import helper


class TestPolicyModel(base.SqlTestCase):
    def setUp(self):
        super(TestPolicyModel, self).setUp()
        # Here we load the fake driver
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        self.engine, self.rule_api, self.policy_model = self.create_services()
        self.initial_policies = set(self.engine.policy_names())
        self._add_test_policy()

    def create_services(self):
        if cfg.CONF.distributed_architecture:
            messaging_config = helper.generate_messaging_config()

            # TODO(masa): following initializing DseNode will be just a fake
            # until API model and Policy Engine support rpc. After these
            # support rpc, pub/sub and so on, replaced with each class.
            engine = dse_node.DseNode(messaging_config, 'engine', [])
            rule_api = dse_node.DseNode(messaging_config, 'api-rule', [])
            policy_api = dse_node.DseNode(messaging_config,
                                          'policy_model', [])
            for n in (engine, rule_api, policy_api):
                n.start()
        else:
            cage = harness.create(helper.root_path())
            engine = cage.service_object('engine')
            rule_api = cage.service_object('api-rule')
            policy_api = policy_model.PolicyModel("policy_model", {},
                                                  policy_engine=engine)

        return engine, rule_api, policy_api

    def tearDown(self):
        super(TestPolicyModel, self).tearDown()
        if cfg.CONF.distributed_architecture:
            for n in (self.engine, self.rule_api, self.policy_model):
                n.stop()
                n.wait()

    def _add_test_policy(self):
        test_policy = {
            "name": "test-policy",
            "description": "test policy description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }

        test_policy_id, obj = self.policy_model.add_item(test_policy, {})
        test_policy["id"] = test_policy_id
        test_policy["owner_id"] = obj["owner_id"]

        test_policy2 = {
            "name": "test-policy2",
            "description": "test policy2 description",
            "kind": "nonrecursive",
            "abbreviation": "abbr2"
        }

        test_policy_id, obj = self.policy_model.add_item(test_policy2, {})
        test_policy2["id"] = test_policy_id
        test_policy2["owner_id"] = obj["owner_id"]

        self.policy = test_policy
        self.policy2 = test_policy2

    def test_in_mem_and_db_policies(self):
        ret = self.policy_model.get_items({})
        db = [p['name'] for p in ret['results']]
        mem = self.engine.policy_names()
        new_memory = set(mem) - self.initial_policies
        new_db = set(db) - self.initial_policies
        self.assertEqual(new_memory, new_db)

    def test_get_items(self):
        ret = self.policy_model.get_items({})
        self.assertTrue(all(p in ret['results']
                            for p in [self.policy, self.policy2]))

    def test_get_item(self):
        expected_ret = self.policy
        ret = self.policy_model.get_item(self.policy["id"], {})
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_item(self):
        self.assertRaises(KeyError, self.policy_model.get_item,
                          'invalid-id', {})

    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    def test_add_item(self, patched_gen_uuid):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }
        patched_gen_uuid.return_value = 'uuid'
        uuidutils.generate_uuid = mock.Mock()
        uuidutils.generate_uuid.return_value = 'uuid'
        expected_ret1 = 'uuid'
        expected_ret2 = {
            'id': 'uuid',
            'name': test['name'],
            'owner_id': 'user',
            'description': test['description'],
            'abbreviation': test['abbreviation'],
            'kind': test['kind']
        }

        policy_id, policy_obj = self.policy_model.add_item(test, {})
        self.assertEqual(expected_ret1, policy_id)
        self.assertEqual(expected_ret2, policy_obj)

    def test_add_item_with_id(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }

        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item, test, {}, 'id')

    def test_add_item_without_name(self):
        test = {
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }

        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item, test, {})

    def test_add_item_with_long_abbreviation(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "123456"
        }
        try:
            self.policy_model.add_item(test, {})
            self.fail("DataModelException should been raised.")
        except webservice.DataModelException as e:
            error_key = 'policy_abbreviation_error'
            self.assertEqual(error_codes.get_num(error_key), e.error_code)
            self.assertEqual(error_codes.get_desc(error_key), e.description)
            self.assertEqual(error_codes.get_http(error_key),
                             e.http_status_code)

    def test_delete_item(self):
        expected_ret = self.policy
        policy_id = self.policy['id']

        ret = self.policy_model.delete_item(policy_id, {})
        self.assertEqual(expected_ret, ret)
        self.assertRaises(KeyError, self.policy_model.get_item,
                          self.policy['id'], {})

        # check that deleting the policy also deletes the rules
        self.assertRaises(webservice.DataModelException,
                          self.rule_api.get_items,
                          {}, {'policy_id': policy_id})

    def test_simulate_action(self):
        context = {
            'policy_id': self.engine.ACTION_THEORY
        }
        action_rule1 = {
            'rule': 'action("q")',
        }
        action_rule2 = {
            'rule': 'p+(x):- q(x)'
        }
        self.rule_api.add_item(action_rule1, {}, context=context)
        self.rule_api.add_item(action_rule2, {}, context=context)

        request_body = {
            'query': 'p(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'q(1)'
        }
        request = helper.FakeRequest(request_body)
        expected_ret = {
            'result': [
                "p(1)"
            ]
        }

        ret = self.policy_model.simulate_action({}, context, request)
        self.assertEqual(expected_ret, ret)

    def test_simulate_with_delta(self):
        context = {
            'policy_id': self.engine.ACTION_THEORY
        }
        action_rule1 = {
            'rule': 'action("q")',
        }
        action_rule2 = {
            'rule': 'p+(x):- q(x)'
        }
        self.rule_api.add_item(action_rule1, {}, context=context)
        self.rule_api.add_item(action_rule2, {}, context=context)

        request_body = {
            'query': 'p(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'q(1)'
        }
        request = helper.FakeRequest(request_body)
        params = {
            'delta': 'true'
        }
        expected_ret = {
            'result': [
                "p+(1)"
            ]
        }

        ret = self.policy_model.simulate_action(params, context, request)
        self.assertEqual(expected_ret, ret)

    def test_simulate_with_trace(self):
        context = {
            'policy_id': self.engine.ACTION_THEORY
        }
        action_rule1 = {
            'rule': 'action("q")',
        }
        action_rule2 = {
            'rule': 'p+(x):- q(x)'
        }
        self.rule_api.add_item(action_rule1, {}, context=context)
        self.rule_api.add_item(action_rule2, {}, context=context)

        request_body = {
            'query': 'p(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'q(1)'
        }
        request = helper.FakeRequest(request_body)
        params = {
            'trace': 'true'
        }
        expected_ret = {
            'result': [
                "p(1)"
            ],
            'trace': "trace strings"
        }

        ret = self.policy_model.simulate_action(params, context, request)
        # check response's keys equal expected_ret's key
        self.assertTrue(all(key in expected_ret.keys() for key in ret.keys()))
        self.assertEqual(expected_ret['result'], ret['result'])
        self.assertTrue(len(ret['trace']) > 10)

    def test_simulate_with_delta_and_trace(self):
        context = {
            'policy_id': self.engine.ACTION_THEORY
        }
        action_rule1 = {
            'rule': 'action("q")',
        }
        action_rule2 = {
            'rule': 'p+(x):- q(x)'
        }
        self.rule_api.add_item(action_rule1, {}, context=context)
        self.rule_api.add_item(action_rule2, {}, context=context)

        request_body = {
            'query': 'p(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'q(1)'
        }
        request = helper.FakeRequest(request_body)
        params = {
            'trace': 'true',
            'delta': 'true'
        }
        expected_ret = {
            'result': [
                "p+(1)"
            ],
            'trace': "trace strings"
        }

        ret = self.policy_model.simulate_action(params, context, request)
        # check response's keys equal expected_ret's key
        self.assertTrue(all(key in expected_ret.keys() for key in ret.keys()))
        self.assertEqual(expected_ret['result'], ret['result'])
        self.assertTrue(len(ret['trace']) > 10)

    def test_simulate_invalid_policy(self):
        context = {
            'policy_id': 'invalid-policy'
        }
        request_body = {
            'query': 'p(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'q(1)'
        }
        request = helper.FakeRequest(request_body)

        self.assertRaises(webservice.DataModelException,
                          self.policy_model.simulate_action,
                          {}, context, request)

    def test_simulate_invalid_sequence(self):
        context = {
            'policy_id': self.engine.ACTION_THEORY
        }
        action_rule = {
            'rule': 'w(x):-z(x)',
        }
        self.rule_api.add_item(action_rule, {}, context=context)

        request_body = {
            'query': 'w(x)',
            'action_policy': self.engine.ACTION_THEORY,
            'sequence': 'z(1)'
        }
        request = helper.FakeRequest(request_body)

        self.assertRaises(webservice.DataModelException,
                          self.policy_model.simulate_action,
                          {}, context, request)

    def test_simulate_policy_errors(self):
        def check_err(params, context, request, emsg):
            try:
                self.policy_model.simulate_action(params, context, request)
                self.assertFail()
            except webservice.DataModelException as e:
                self.assertIn(emsg, str(e))

        context = {
            'policy_id': self.engine.ACTION_THEORY
        }

        # Missing query
        body = {'action_policy': self.engine.ACTION_THEORY,
                'sequence': 'q(1)'}
        check_err({}, context, helper.FakeRequest(body),
                  'Simulate requires parameters')

        # Invalid query
        body = {'query': 'p(x',
                'action_policy': self.engine.ACTION_THEORY,
                'sequence': 'q(1)'}
        check_err({}, context, helper.FakeRequest(body),
                  'Parse failure')

        # Multiple querys
        body = {'query': 'p(x) q(x)',
                'action_policy': self.engine.ACTION_THEORY,
                'sequence': 'q(1)'}
        check_err({}, context, helper.FakeRequest(body),
                  'more than 1 rule')

        # Missing action_policy
        body = {'query': 'p(x)',
                'sequence': 'q(1)'}
        check_err({}, context, helper.FakeRequest(body),
                  'Simulate requires parameters')

        # Missing sequence
        body = {'query': 'p(x)',
                'action_policy': self.engine.ACTION_THEORY}
        check_err({}, context, helper.FakeRequest(body),
                  'Simulate requires parameters')

        # Syntactically invalid sequence
        body = {'query': 'p(x)',
                'action_policy': self.engine.ACTION_THEORY,
                'sequence': 'q(1'}
        check_err({}, context, helper.FakeRequest(body),
                  'Parse failure')

        # Semantically invalid sequence
        body = {'query': 'p(x)',
                'action_policy': self.engine.ACTION_THEORY,
                'sequence': 'r(1)'}  # r is not an action
        check_err({}, context, helper.FakeRequest(body),
                  'non-action, non-update')

    def test_policy_api_model_error(self):
        """Test the policy api model."""

        # add policy without name
        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item, {}, {})

        # add policy with bad ID
        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item, {'name': '7*7'}, {})
        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item,
                          {'name': 'p(x) :- q(x)'}, {})

        # add policy with invalid 'kind'
        self.assertRaises(webservice.DataModelException,
                          self.policy_model.add_item,
                          {'kind': 'nonexistent', 'name': 'alice'}, {})

        # add existing policy
        self.policy_model.add_item({'name': 'Test1'}, {})
        self.assertRaises(KeyError, self.policy_model.add_item,
                          {'name': 'Test1'}, {})

        # delete non-existent policy
        self.assertRaises(KeyError, self.policy_model.delete_item,
                          'noexist', {})

        # delete system-maintained policy
        policies = self.policy_model.get_items({})['results']
        class_policy = [p for p in policies if p['name'] == 'classification']
        class_policy = class_policy[0]
        self.assertRaises(KeyError, self.policy_model.delete_item,
                          class_policy['id'], {})



# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
from oslo_db.sqlalchemy import session

_FACADE = None


def _create_facade_lazily():
    global _FACADE

    if _FACADE is None:
        _FACADE = session.EngineFacade.from_config(cfg.CONF, sqlite_fk=True)

    return _FACADE


def get_engine():
    """Helper method to grab engine."""
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(autocommit=True, expire_on_commit=False):
    """Helper method to grab session."""
    facade = _create_facade_lazily()
    return facade.get_session(autocommit=autocommit,
                              expire_on_commit=expire_on_commit)



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import compile
from congress.datalog import topdown
from congress.datalog import unify
from congress.tests import base

LOG = logging.getLogger(__name__)


NREC_THEORY = 'non-recursive theory'
DB_THEORY = 'database'
MAT_THEORY = 'materialized'


class TestUnify(base.TestCase):
    def open(self, msg):
        LOG.debug("** Started: %s **", msg)

    def close(self, msg):
        LOG.debug("** Finished: %s **", msg)

    def create_unify(self, atom_string1, atom_string2, msg, change_num,
                     unifier1=None, unifier2=None, recursive_str=False):
        """Create unification and check basic results."""
        def str_uni(u):
            if recursive_str:
                return u.recur_str()
            else:
                return str(u)

        def print_unifiers(changes=None):
            LOG.debug("unifier1: %s", str_uni(unifier1))
            LOG.debug("unifier2: %s", str_uni(unifier2))
            if changes is not None:
                LOG.debug("changes: %s",
                          ";".join([str(x) for x in changes]))

        if msg is not None:
            self.open(msg)
        if unifier1 is None:
            # LOG.debug("Generating new unifier1")
            unifier1 = topdown.TopDownTheory.new_bi_unifier()
        if unifier2 is None:
            # LOG.debug("Generating new unifier2")
            unifier2 = topdown.TopDownTheory.new_bi_unifier()
        p1 = compile.parse(atom_string1)[0]
        p2 = compile.parse(atom_string2)[0]
        changes = unify.bi_unify_atoms(p1, unifier1, p2, unifier2)
        self.assertTrue(changes is not None)
        print_unifiers(changes)
        p1p = p1.plug(unifier1)
        p2p = p2.plug(unifier2)
        print_unifiers(changes)
        if not p1p == p2p:
            LOG.debug("Failure: bi-unify(%s, %s) produced %s and %s",
                      p1, p2, str_uni(unifier1), str_uni(unifier2))
            LOG.debug("plug(%s, %s) = %s", p1, str_uni(unifier1), p1p)
            LOG.debug("plug(%s, %s) = %s", p2, str_uni(unifier2), p2p)
            self.fail()
        if change_num is not None and len(changes) != change_num:
            LOG.debug("Failure: bi-unify(%s, %s) produced %s and %s",
                      p1, p2, str_uni(unifier1), str_uni(unifier2))
            LOG.debug("plug(%s, %s) = %s", p1, str_uni(unifier1), p1p)
            LOG.debug("plug(%s, %s) = %s", p2, str_uni(unifier2), p2p)
            LOG.debug("Expected %s changes; computed %s changes",
                      change_num, len(changes))
            self.fail()
        LOG.debug("unifier1: %s", str_uni(unifier1))
        LOG.debug("unifier2: %s", str_uni(unifier2))
        if msg is not None:
            self.open(msg)
        return (p1, unifier1, p2, unifier2, changes)

    def check_unify(self, atom_string1, atom_string2, msg, change_num,
                    unifier1=None, unifier2=None, recursive_str=False):
        self.open(msg)
        (p1, unifier1, p2, unifier2, changes) = self.create_unify(
            atom_string1, atom_string2, msg, change_num,
            unifier1=unifier1, unifier2=unifier2, recursive_str=recursive_str)
        unify.undo_all(changes)
        self.assertEqual(p1, p1.plug(unifier1))
        self.assertEqual(p2, p2.plug(unifier2))
        self.close(msg)

    def check_unify_fail(self, atom_string1, atom_string2, msg):
        """Check that the bi-unification fails."""
        self.open(msg)
        unifier1 = topdown.TopDownTheory.new_bi_unifier()
        unifier2 = topdown.TopDownTheory.new_bi_unifier()
        p1 = compile.parse(atom_string1)[0]
        p2 = compile.parse(atom_string2)[0]
        changes = unify.bi_unify_atoms(p1, unifier1, p2, unifier2)
        if changes is not None:
            LOG.debug("Failure failure: bi-unify(%s, %s) produced %s and %s",
                      p1, p2, unifier1, unifier2)
            LOG.debug("plug(%s, %s) = %s", p1, unifier1, p1.plug(unifier1))
            LOG.debug("plug(%s, %s) = %s", p2, unifier2, p2.plug(unifier2))
            self.fail()
        self.close(msg)

    def test_instance(self):
        """Test whether the INSTANCE computation is correct."""
        def assertIsNotNone(x):
            self.assertTrue(x is not None)

        def assertIsNone(x):
            self.assertTrue(x is None)

        assertIsNotNone(unify.instance(str2form('p(1)'), str2form('p(y)')))
        assertIsNotNone(unify.instance(str2form('p(1,2)'), str2form('p(x,y)')))
        assertIsNotNone(unify.instance(str2form('p(1,x)'), str2form('p(x,y)')))
        assertIsNotNone(unify.instance(str2form('p(1,x,1)'),
                        str2form('p(x,y,x)')))
        assertIsNotNone(unify.instance(str2form('p(1,x,1)'),
                        str2form('p(x,y,z)')))
        assertIsNone(unify.instance(str2form('p(1,2)'), str2form('p(x,x)')))

    def test_same(self):
        """Test whether the SAME computation is correct."""
        def assertIsNotNone(x):
            self.assertTrue(x is not None)

        def assertIsNone(x):
            self.assertTrue(x is None)

        assertIsNotNone(unify.same(str2form('p(x)'), str2form('p(y)')))
        assertIsNotNone(unify.same(str2form('p(x)'), str2form('p(x)')))
        assertIsNotNone(unify.same(str2form('p(x,y)'), str2form('p(x,y)')))
        assertIsNotNone(unify.same(str2form('p(x,y)'), str2form('p(y,x)')))
        assertIsNone(unify.same(str2form('p(x,x)'), str2form('p(x,y)')))
        assertIsNone(unify.same(str2form('p(x,y)'), str2form('p(x,x)')))
        assertIsNotNone(unify.same(str2form('p(x,x)'), str2form('p(y,y)')))
        assertIsNotNone(unify.same(str2form('p(x,y,x)'), str2form('p(y,x,y)')))
        assertIsNone(unify.same(str2form('p(x,y,z)'), str2form('p(x,y,1)')))

    def test_bi_unify(self):
        """Test the bi-unification routine and its supporting routines."""
        def var(x):
            return compile.Term.create_from_python(x, force_var=True)

        def obj(x):
            return compile.Term.create_from_python(x)

        def new_uni():
            return topdown.TopDownTheory.new_bi_unifier()

        # apply, add
        u1 = new_uni()
        u1.add(var('x'), obj(1), None)
        self.assertEqual(u1.apply(var('x')), obj(1))

        u1 = new_uni()
        u2 = new_uni()
        u1.add(var('y'), var('x'), u2)
        self.assertEqual(u1.apply(var('y')), var('x'))
        u2.add(var('x'), obj(2), None)
        self.assertEqual(u1.apply(var('y')), obj(2))

        # delete
        u1.delete(var('y'))
        self.assertEqual(u1.apply(var('y')), var('y'))

        u1 = new_uni()
        u2 = new_uni()
        u1.add(var('y'), var('x'), u2)
        u2.add(var('x'), obj(2), None)
        u2.delete(var('x'))
        self.assertEqual(u1.apply(var('y')), var('x'))
        u1.delete(var('y'))
        self.assertEqual(u1.apply(var('y')), var('y'))

        # bi_unify
        self.check_unify("p(x)", "p(1)",
                         "Matching", 1)
        self.check_unify("p(x,y)", "p(1,2)",
                         "Binary Matching", 2)
        self.check_unify("p(1,2)", "p(x,y)",
                         "Binary Matching Reversed", 2)
        self.check_unify("p(1,1)", "p(x,y)",
                         "Binary Matching Many-to-1", 2)
        self.check_unify_fail("p(1,2)", "p(x,x)",
                              "Binary Matching Failure")
        self.check_unify("p(1,x)", "p(1,y)",
                         "Simple Unification", 1)
        self.check_unify("p(1,x)", "p(y,1)",
                         "Separate Namespace Unification", 2)
        self.check_unify("p(1,x)", "p(x,2)",
                         "Namespace Collision Unification", 2)
        self.check_unify("p(x,y,z)", "p(t,u,v)",
                         "Variable-only Unification", 3)
        self.check_unify("p(x,y,y)", "p(t,u,t)",
                         "Repeated Variable Unification", 3)
        self.check_unify_fail("p(x,y,y,x,y)", "p(t,u,t,1,2)",
                              "Repeated Variable Unification Failure")
        self.check_unify(
            "p(x,y,y)", "p(x,y,x)",
            "Repeated Variable Unification Namespace Collision", 3)
        self.check_unify_fail(
            "p(x,y,y,x,y)", "p(x,y,x,1,2)",
            "Repeated Variable Unification Namespace Collision Failure")

        # test sequence of changes
        (p1, u1, p2, u2, changes) = self.create_unify(
            "p(x)", "p(x)", "Step 1", 1)   # 1 since the two xs are different
        self.create_unify(
            "p(x)", "p(1)", "Step 2", 1, unifier1=u1, recursive_str=True)
        self.create_unify(
            "p(x)", "p(1)", "Step 3", 0, unifier1=u1, recursive_str=True)


class TestMatch(base.TestCase):

    def check(self, atom1, atom2):
        atom1 = compile.parse1(atom1)
        atom2 = compile.parse1(atom2)
        unifier = unify.BiUnifier()
        changes = unify.match_atoms(atom1, unifier, atom2)
        self.assertIsNotNone(changes)
        self.assertEqual(atom1.plug(unifier), atom2)

    def cherr(self, atom1, atom2):
        atom1 = compile.parse1(atom1)
        atom2 = compile.parse1(atom2)
        unifier = unify.BiUnifier()
        self.assertIsNone(unify.match_atoms(atom1, unifier, atom2))

    def test_atoms(self):
        self.check('p(x, y)', 'p(1, 2)')
        self.check('p(x, x)', 'p(1, 1)')
        self.cherr('p(x, x)', 'p(1, 2)')
        self.check('p(x, y, z)', 'p(1, 2, 1)')
        self.check('p(x, y, x, z)', 'p(1, 2, 1, 3)')
        self.cherr('p(x, y, x, y)', 'p(1, 2, 1, 3)')
        self.cherr('p(x, y, x, y)', 'p(1, 2, 2, 1)')
        self.check('p(x, y, x, y)', 'p(1, 1, 1, 1)')

    def test_sequence(self):
        atom1 = compile.parse1('p(x, y)')
        atom2 = compile.parse1('p(1, 2)')
        unifier = unify.BiUnifier()
        changes = unify.match_atoms(atom1, unifier, atom2)
        self.assertIsNotNone(changes)

        atom3 = compile.parse1('q(y, z)')
        atom4 = compile.parse1('q(2, 3)')
        changes = unify.match_atoms(atom3, unifier, atom4)
        self.assertIsNotNone(changes)

        atom5 = compile.parse1('r(x, y, z, z)')
        atom6 = compile.parse1('r(1, 2, 3, 3)')
        changes = unify.match_atoms(atom5, unifier, atom6)
        self.assertIsNotNone(changes)

        self.assertEqual(atom1.plug(unifier), atom2)
        self.assertEqual(atom3.plug(unifier), atom4)
        self.assertEqual(atom5.plug(unifier), atom6)


def str2form(formula_string):
    return compile.parse1(formula_string)


def str2pol(policy_string):
    return compile.parse(policy_string)


def pol2str(policy):
    return " ".join(str(x) for x in policy)


def form2str(formula):
    return str(formula)



# Copyright (c) 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_config import cfg

from congress.policy_engines.agnostic import Dse2Runtime
from congress.tests import fake_datasource
from congress.tests import helper


def setup_config(services=[]):
    cfg.CONF.set_override('distributed_architecture', True)
    # Load the fake driver.
    cfg.CONF.set_override(
        'drivers',
        ['congress.tests.fake_datasource.FakeDataSource'])

    node = helper.make_dsenode_new_partition("testnode")
    engine = Dse2Runtime('engine')
    data = fake_datasource.FakeDataSource('data')

    node.register_service(engine)
    node.register_service(data)

    for service in services:
        node.register_service(service)

    return {'node': node, 'engine': engine, 'data': data}



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging
from six.moves import range

from congress.datalog import base
from congress.datalog import compile
from congress.datalog import database
from congress.datalog import topdown
from congress.datalog import utility


LOG = logging.getLogger(__name__)


class DeltaRule(object):
    """Rule describing how updates to data sources change table."""
    def __init__(self, trigger, head, body, original):
        self.trigger = trigger  # atom
        self.head = head  # atom

        # list of literals, sorted for order-insensitive comparison
        self.body = (
            sorted([lit for lit in body if not lit.is_builtin()]) +
            sorted([lit for lit in body if lit.is_builtin()]))
        self.original = original  # Rule from which SELF was derived

    def __str__(self):
        return "<trigger: {}, head: {}, body: {}>".format(
            str(self.trigger), str(self.head), [str(lit) for lit in self.body])

    def __eq__(self, other):
        return (self.trigger == other.trigger and
                self.head == other.head and
                len(self.body) == len(other.body) and
                all(self.body[i] == other.body[i]
                    for i in range(0, len(self.body))))

    def __hash__(self):
        return hash((self.trigger, self.head, tuple(self.body)))

    def variables(self):
        """Return the set of variables occurring in this delta rule."""
        vs = self.trigger.variables()
        vs |= self.head.variables()
        for atom in self.body:
            vs |= atom.variables()
        return vs

    def tablenames(self, body_only=False, include_builtin=False,
                   include_modal=True):
        """Return the set of tablenames occurring in this delta rule."""
        tables = set()
        if not body_only:
            tables.add(self.head.tablename())
        tables.add(self.trigger.tablename())
        for atom in self.body:
            tables.add(atom.tablename())
        return tables


class DeltaRuleTheory (base.Theory):
    """A collection of DeltaRules.  Not useful by itself as a policy."""
    def __init__(self, name=None, abbr=None, theories=None):
        super(DeltaRuleTheory, self).__init__(
            name=name, abbr=abbr, theories=theories)
        # dictionary from table name to list of rules with that table as
        # trigger
        self.rules = {}
        # dictionary from delta_rule to the rule from which it was derived
        self.originals = set()
        # dictionary from table name to number of rules with that table in
        # head
        self.views = {}
        # all tables
        self.all_tables = {}
        self.kind = base.DELTA_POLICY_TYPE

    def modify(self, event):
        """Insert/delete the compile.Rule RULE into the theory.

        Return list of changes (either the empty list or
        a list including just RULE).
        """
        self.log(None, "DeltaRuleTheory.modify %s", event.formula)
        self.log(None, "originals: %s", utility.iterstr(self.originals))
        if event.insert:
            if self.insert(event.formula):
                return [event]
        else:
            if self.delete(event.formula):
                return [event]
        return []

    def insert(self, rule):
        """Insert a compile.Rule into the theory.

        Return True iff the theory changed.
        """
        assert compile.is_regular_rule(rule), (
            "DeltaRuleTheory only takes rules")
        self.log(rule.tablename(), "Insert: %s", rule)
        if rule in self.originals:
            self.log(None, utility.iterstr(self.originals))
            return False
        self.log(rule.tablename(), "Insert 2: %s", rule)
        for delta in self.compute_delta_rules([rule]):
            self.insert_delta(delta)
        self.originals.add(rule)
        return True

    def insert_delta(self, delta):
        """Insert a delta rule."""
        self.log(None, "Inserting delta rule %s", delta)
        # views (tables occurring in head)
        if delta.head.table.table in self.views:
            self.views[delta.head.table.table] += 1
        else:
            self.views[delta.head.table.table] = 1

        # tables
        for table in delta.tablenames():
            if table in self.all_tables:
                self.all_tables[table] += 1
            else:
                self.all_tables[table] = 1

        # contents
        if delta.trigger.table.table not in self.rules:
            self.rules[delta.trigger.table.table] = utility.OrderedSet()
        self.rules[delta.trigger.table.table].add(delta)

    def delete(self, rule):
        """Delete a compile.Rule from theory.

        Assumes that COMPUTE_DELTA_RULES is deterministic.
        Returns True iff the theory changed.
        """
        self.log(rule.tablename(), "Delete: %s", rule)
        if rule not in self.originals:
            return False
        for delta in self.compute_delta_rules([rule]):
            self.delete_delta(delta)
        self.originals.remove(rule)
        return True

    def delete_delta(self, delta):
        """Delete the DeltaRule DELTA from the theory."""
        # views
        if delta.head.table.table in self.views:
            self.views[delta.head.table.table] -= 1
            if self.views[delta.head.table.table] == 0:
                del self.views[delta.head.table.table]

        # tables
        for table in delta.tablenames():
            if table in self.all_tables:
                self.all_tables[table] -= 1
                if self.all_tables[table] == 0:
                    del self.all_tables[table]

        # contents
        self.rules[delta.trigger.table.table].discard(delta)
        if not len(self.rules[delta.trigger.table.table]):
            del self.rules[delta.trigger.table.table]

    def policy(self):
        return self.originals

    def get_arity_self(self, tablename):
        for p in self.originals:
            if p.head.table.table == tablename:
                return len(p.head.arguments)
        return None

    def __contains__(self, formula):
        return formula in self.originals

    def __str__(self):
        return str(self.rules)

    def rules_with_trigger(self, table):
        """Return the list of DeltaRules that trigger on the given TABLE."""
        if table in self.rules:
            return self.rules[table]
        else:
            return []

    def is_view(self, x):
        return x in self.views

    def is_known(self, x):
        return x in self.all_tables

    def base_tables(self):
        base = []
        for table in self.all_tables:
            if table not in self.views:
                base.append(table)
        return base

    @classmethod
    def eliminate_self_joins(cls, formulas):
        """Remove self joins.

        Return new list of formulas that is equivalent to
        the list of formulas FORMULAS except that there
        are no self-joins.
        """
        def new_table_name(name, arity, index):
            return "___{}_{}_{}".format(name, arity, index)

        def n_variables(n):
            vars = []
            for i in range(0, n):
                vars.append("x" + str(i))
            return vars
        # dict from (table name, arity) tuple to
        #      max num of occurrences of self-joins in any rule
        global_self_joins = {}
        # remove self-joins from rules
        results = []
        for rule in formulas:
            if rule.is_atom():
                results.append(rule)
                continue
            LOG.debug("eliminating self joins from %s", rule)
            occurrences = {}  # for just this rule
            for atom in rule.body:
                table = atom.tablename()
                arity = len(atom.arguments)
                tablearity = (table, arity)
                if tablearity not in occurrences:
                    occurrences[tablearity] = 1
                else:
                    # change name of atom
                    atom.table.table = new_table_name(table, arity,
                                                      occurrences[tablearity])
                    # update our counters
                    occurrences[tablearity] += 1
                    if tablearity not in global_self_joins:
                        global_self_joins[tablearity] = 1
                    else:
                        global_self_joins[tablearity] = (
                            max(occurrences[tablearity] - 1,
                                global_self_joins[tablearity]))
            results.append(rule)
            LOG.debug("final rule: %s", rule)
        # add definitions for new tables
        for tablearity in global_self_joins:
            table = tablearity[0]
            arity = tablearity[1]
            for i in range(1, global_self_joins[tablearity] + 1):
                newtable = new_table_name(table, arity, i)
                args = [compile.Variable(var) for var in n_variables(arity)]
                head = compile.Literal(newtable, args)
                body = [compile.Literal(table, args)]
                results.append(compile.Rule(head, body))
                LOG.debug("Adding rule %s", results[-1])
        return results

    @classmethod
    def compute_delta_rules(cls, formulas):
        """Return list of DeltaRules computed from formulas.

        Assuming FORMULAS has no self-joins, return a list of DeltaRules
        derived from those FORMULAS.
        """
        # Should do the following for correctness, but it needs to be
        #    done elsewhere so that we can properly maintain the tables
        #    that are generated.
        # formulas = cls.eliminate_self_joins(formulas)
        delta_rules = []
        for rule in formulas:
            if rule.is_atom():
                continue
            rule = compile.reorder_for_safety(rule)
            for literal in rule.body:
                if literal.is_builtin():
                    continue
                newbody = [lit for lit in rule.body if lit is not literal]
                delta_rules.append(
                    DeltaRule(literal, rule.head, newbody, rule))
        return delta_rules


class MaterializedViewTheory(topdown.TopDownTheory):
    """A theory that stores the table contents of views explicitly.

    Relies on included theories to define the contents of those
    tables not defined by the rules of the theory.
    Recursive rules are allowed.
    """

    def __init__(self, name=None, abbr=None, theories=None, schema=None,
                 desc=None, owner=None):
        super(MaterializedViewTheory, self).__init__(
            name=name, abbr=abbr, theories=theories, schema=schema,
            desc=desc, owner=owner)
        # queue of events left to process
        self.queue = base.EventQueue()
        # data storage
        db_name = None
        db_abbr = None
        delta_name = None
        delta_abbr = None
        if name is not None:
            db_name = name + "Database"
            delta_name = name + "Delta"
        if abbr is not None:
            db_abbr = abbr + "DB"
            delta_abbr = abbr + "Dlta"
        self.database = database.Database(name=db_name, abbr=db_abbr)
        # rules that dictate how database changes in response to events
        self.delta_rules = DeltaRuleTheory(name=delta_name, abbr=delta_abbr)
        self.kind = base.MATERIALIZED_POLICY_TYPE

    def set_tracer(self, tracer):
        if isinstance(tracer, base.Tracer):
            self.tracer = tracer
            self.database.tracer = tracer
            self.delta_rules.tracer = tracer
        else:
            self.tracer = tracer['self']
            self.database.tracer = tracer['database']
            self.delta_rules.tracer = tracer['delta_rules']

    def get_tracer(self):
        return {'self': self.tracer,
                'database': self.database.tracer,
                'delta_rules': self.delta_rules.tracer}

    # External Interface

    # SELECT is handled by TopDownTheory

    def insert(self, formula):
        return self.update([compile.Event(formula=formula, insert=True)])

    def delete(self, formula):
        return self.update([compile.Event(formula=formula, insert=False)])

    def update(self, events):
        """Apply inserts/deletes described by EVENTS and return changes.

           Does not check if EVENTS would cause errors.
           """
        for event in events:
            assert compile.is_datalog(event.formula), (
                "Non-formula not allowed: {}".format(str(event.formula)))
            self.enqueue_any(event)
        changes = self.process_queue()
        return changes

    def update_would_cause_errors(self, events):
        """Return a list of PolicyException.

        Return a list of PolicyException if we were
        to apply the events EVENTS to the current policy.
        """
        self.log(None, "update_would_cause_errors %s", utility.iterstr(events))
        errors = []
        # compute new rule set
        for event in events:
            assert compile.is_datalog(event.formula), (
                "update_would_cause_errors operates only on objects")
            self.log(None, "Updating %s", event.formula)
            if event.formula.is_atom():
                errors.extend(compile.fact_errors(
                    event.formula, self.theories, self.name))
            else:
                errors.extend(compile.rule_errors(
                    event.formula, self.theories, self.name))
        return errors

    def explain(self, query, tablenames, find_all):
        """Returns a list of proofs if QUERY is true or None if else."""
        assert compile.is_atom(query), "Explain requires an atom"
        # ignoring TABLENAMES and FIND_ALL
        #    except that we return the proper type.
        proof = self.explain_aux(query, 0)
        if proof is None:
            return None
        else:
            return [proof]

    def policy(self):
        return self.delta_rules.policy()

    def get_arity_self(self, tablename):
        result = self.database.get_arity_self(tablename)
        if result:
            return result
        return self.delta_rules.get_arity_self(tablename)

    # Interface implementation

    def explain_aux(self, query, depth):
        self.log(query.table.table, "Explaining %s", query, depth=depth)
        # Bail out on negated literals.  Need different
        #   algorithm b/c we need to introduce quantifiers.
        if query.is_negated():
            return base.Proof(query, [])
        # grab first local proof, since they're all equally good
        localproofs = self.database.explain(query)
        if localproofs is None:
            return None
        if len(localproofs) == 0:   # base fact
            return base.Proof(query, [])
        localproof = localproofs[0]
        rule_instance = localproof.rule.plug(localproof.binding)
        subproofs = []
        for lit in rule_instance.body:
            subproof = self.explain_aux(lit, depth + 1)
            if subproof is None:
                return None
            subproofs.append(subproof)
        return base.Proof(query, subproofs)

    def modify(self, event):
        """Modifies contents of theory to insert/delete FORMULA.

        Returns True iff the theory changed.
        """
        self.log(None, "Materialized.modify")
        self.enqueue_any(event)
        changes = self.process_queue()
        self.log(event.formula.tablename(),
                 "modify returns %s", utility.iterstr(changes))
        return changes

    def enqueue_any(self, event):
        """Enqueue event.

        Processing rules is a bit different than processing atoms
        in that they generate additional events that we want
        to process either before the rule is deleted or after
        it is inserted.  PROCESS_QUEUE is similar but assumes
        that only the data will cause propagations (and ignores
        included theories).
        """
        # Note: all included theories must define MODIFY
        formula = event.formula
        if formula.is_atom():
            self.log(formula.tablename(), "compute/enq: atom %s", formula)
            assert not self.is_view(formula.table.table), (
                "Cannot directly modify tables" +
                " computed from other tables")
            # self.log(formula.table, "%s: %s", text, formula)
            self.enqueue(event)
            return []
        else:
            # rules do not need to talk to included theories because they
            #   only generate events for views
            # need to eliminate self-joins here so that we fill all
            #   the tables introduced by self-join elimination.
            for rule in DeltaRuleTheory.eliminate_self_joins([formula]):
                new_event = compile.Event(formula=rule, insert=event.insert,
                                          target=event.target)
                self.enqueue(new_event)
            return []

    def enqueue(self, event):
        self.log(event.tablename(), "Enqueueing: %s", event)
        self.queue.enqueue(event)

    def process_queue(self):
        """Data and rule propagation routine.

        Returns list of events that were not noops
        """
        self.log(None, "Processing queue")
        history = []
        while len(self.queue) > 0:
            event = self.queue.dequeue()
            self.log(event.tablename(), "Dequeued %s", event)
            if compile.is_regular_rule(event.formula):
                changes = self.delta_rules.modify(event)
                if len(changes) > 0:
                    history.extend(changes)
                    bindings = self.top_down_evaluation(
                        event.formula.variables(), event.formula.body)
                    self.log(event.formula.tablename(),
                             "new bindings after top-down: %s",
                             utility.iterstr(bindings))
                    self.process_new_bindings(bindings, event.formula.head,
                                              event.insert, event.formula)
            else:
                self.propagate(event)
                history.extend(self.database.modify(event))
            self.log(event.tablename(), "History: %s",
                     utility.iterstr(history))
        return history

    def propagate(self, event):
        """Propagate event.

        Computes and enqueue events generated by EVENT and the DELTA_RULES.
        """
        self.log(event.formula.table.table, "Processing event: %s", event)
        applicable_rules = self.delta_rules.rules_with_trigger(
            event.formula.table.table)
        if len(applicable_rules) == 0:
            self.log(event.formula.table.table, "No applicable delta rule")
        for delta_rule in applicable_rules:
            self.propagate_rule(event, delta_rule)

    def propagate_rule(self, event, delta_rule):
        """Propagate event and delta_rule.

        Compute and enqueue new events generated by EVENT and DELTA_RULE.
        """
        self.log(event.formula.table.table, "Processing event %s with rule %s",
                 event, delta_rule)

        # compute tuples generated by event (either for insert or delete)
        # print "event: {}, event.tuple: {},
        #     event.tuple.rawtuple(): {}".format(
        #     str(event), str(event.tuple), str(event.tuple.raw_tuple()))
        # binding_list is dictionary

        # Save binding for delta_rule.trigger; throw away binding for event
        #   since event is ground.
        binding = self.new_bi_unifier()
        assert compile.is_literal(delta_rule.trigger)
        assert compile.is_literal(event.formula)
        undo = self.bi_unify(delta_rule.trigger, binding,
                             event.formula, self.new_bi_unifier(), self.name)
        if undo is None:
            return
        self.log(event.formula.table.table,
                 "binding list for event and delta-rule trigger: %s", binding)
        bindings = self.top_down_evaluation(
            delta_rule.variables(), delta_rule.body, binding)
        self.log(event.formula.table.table, "new bindings after top-down: %s",
                 ",".join([str(x) for x in bindings]))

        if delta_rule.trigger.is_negated():
            insert_delete = not event.insert
        else:
            insert_delete = event.insert
        self.process_new_bindings(bindings, delta_rule.head,
                                  insert_delete, delta_rule.original)

    def process_new_bindings(self, bindings, atom, insert, original_rule):
        """Process new bindings.

        For each of BINDINGS, apply to ATOM, and enqueue it as an insert if
        INSERT is True and as a delete otherwise.
        """
        # for each binding, compute generated tuple and group bindings
        #    by the tuple they generated
        new_atoms = {}
        for binding in bindings:
            new_atom = atom.plug(binding)
            if new_atom not in new_atoms:
                new_atoms[new_atom] = []
            new_atoms[new_atom].append(database.Database.Proof(
                binding, original_rule))
        self.log(atom.table.table, "new tuples generated: %s",
                 utility.iterstr(new_atoms))

        # enqueue each distinct generated tuple, recording appropriate bindings
        for new_atom in new_atoms:
            # self.log(event.table, "new_tuple %s: %s", new_tuple,
            #          new_tuples[new_tuple])
            # Only enqueue if new data.
            # Putting the check here is necessary to support recursion.
            self.enqueue(compile.Event(formula=new_atom,
                         proofs=new_atoms[new_atom],
                         insert=insert))

    def is_view(self, x):
        """Return True if the table X is defined by the theory."""
        return self.delta_rules.is_view(x)

    def is_known(self, x):
        """Return True if this theory has any rule mentioning table X."""
        return self.delta_rules.is_known(x)

    def base_tables(self):
        """Get base tables.

        Return the list of tables that are mentioned in the rules but
        for which there are no rules with those tables in the head.
        """
        return self.delta_rules.base_tables()

    def _top_down_th(self, context, caller):
        return self.database._top_down_th(context, caller)

    def content(self, tablenames=None):
        return self.database.content(tablenames=tablenames)

    def __contains__(self, formula):
        # TODO(thinrichs): if formula is a rule, we need to check
        #   self.delta_rules; if formula is an atom, we need to check
        #   self.database, but only if the table for that atom is
        #   not defined by rules.  As it stands, for atoms, we are
        #   conflating membership with evaluation.
        return (formula in self.database or formula in self.delta_rules)



# Copyright (c) 2014 Marist SDN Innovation lab Joint with Plexxi Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging
from oslo_vmware import api
from oslo_vmware import vim_util

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance.

    """

    return VCenterDriver(name, keys, inbox, datapath, args)


class VCenterDriver(datasource_driver.PollingDataSourceDriver,
                    datasource_driver.ExecutionDriver):

    HOSTS = "hosts"
    HOST_DNS = "host.DNS_IPs"
    HOST_PNICS = "host.PNICs"
    HOST_VNICS = "host.VNICs"
    VMS = "vms"

    value_trans = {'translation-type': 'VALUE'}

    vms_translator = {
        'translation-type': 'HDICT',
        'table-name': VMS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'uuid', 'translator': value_trans},
             {'fieldname': 'host_uuid', 'translator': value_trans},
             {'fieldname': 'pathName', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'CpuDemand', 'translator': value_trans},
             {'fieldname': 'CpuUsage', 'translator': value_trans},
             {'fieldname': 'memorySizeMB', 'translator': value_trans},
             {'fieldname': 'MemoryUsage', 'translator': value_trans},
             {'fieldname': 'committedStorage', 'translator': value_trans},
             {'fieldname': 'uncommittedStorage', 'translator': value_trans},
             {'fieldname': 'annotation', 'translator': value_trans})}

    pnic_translator = {
        'translation-type': 'HDICT',
        'table-name': HOST_PNICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'host_uuid', 'translator': value_trans},
             {'fieldname': 'device', 'translator': value_trans},
             {'fieldname': 'mac', 'translator': value_trans},
             {'fieldname': 'ipAddress', 'translator': value_trans},
             {'fieldname': 'subnetMask', 'translator': value_trans})}

    vnic_translator = {
        'translation-type': 'HDICT',
        'table-name': HOST_VNICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'host_uuid', 'translator': value_trans},
             {'fieldname': 'device', 'translator': value_trans},
             {'fieldname': 'mac', 'translator': value_trans},
             {'fieldname': 'portgroup', 'translator': value_trans},
             {'fieldname': 'ipAddress', 'translator': value_trans},
             {'fieldname': 'subnetMask', 'translator': value_trans})}

    hosts_translator = {
        'translation-type': 'HDICT',
        'table-name': HOSTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'uuid', 'translator': value_trans},
             {'fieldname': HOST_DNS, 'col': 'Host:DNS_id',
              'translator': {'translation-type': 'LIST',
                             'table-name': HOST_DNS,
                             'id-col': 'Host:DNS_id',
                             'val-col': 'DNS_IPs',
                             'translator': value_trans}})}

    TRANSLATORS = [hosts_translator, pnic_translator, vnic_translator,
                   vms_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None,
                 session=None):
        if args is None:
            args = self.empty_credentials()
        else:
            args['tenant_name'] = None
        super(VCenterDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        try:
            self.max_VMs = int(args['max_vms'])
        except (KeyError, ValueError):
            LOG.warning("max_vms has not been configured, "
                        " defaulting to 999.")
            self.max_VMs = 999
        try:
            self.max_Hosts = int(args['max_hosts'])
        except (KeyError, ValueError):
            LOG.warning("max_hosts has not been configured, "
                        "defaulting to 999.")
            self.max_Hosts = 999
        self.hosts = None
        self.creds = args
        self.session = session
        if session is None:
            self.session = api.VMwareAPISession(self.creds['auth_url'],
                                                self.creds['username'],
                                                self.creds['password'],
                                                10, 1,
                                                create_session=True)
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'vcenter'
        result['description'] = ('Datasource driver that interfaces with '
                                 'vcenter')
        result['config'] = {'auth_url': constants.REQUIRED,
                            'username': constants.REQUIRED,
                            'password': constants.REQUIRED,
                            'poll_time': constants.OPTIONAL,
                            'max_vms': constants.OPTIONAL,
                            'max_hosts': constants.OPTIONAL}
        result['secret'] = ['password']

        return result

    def update_from_datasource(self):
        """Called when it is time to pull new data from this datasource.

        Pulls lists of objects from vCenter, if the data does not match
        the correspondig table in the driver's raw state or has not yet been
        added to the state, the driver calls methods to parse this data.
        """

        hosts, pnics, vnics = self._get_hosts_and_nics()
        self._translate_hosts(hosts)
        self._translate_pnics(pnics)
        self._translate_vnics(vnics)

        vms = self._get_vms()
        self._translate_vms(vms)

    @ds_utils.update_state_on_changed(HOSTS)
    def _translate_hosts(self, hosts):
        """Translate the host data from vCenter."""

        row_data = VCenterDriver.convert_objs(hosts,
                                              VCenterDriver.hosts_translator)
        return row_data

    @ds_utils.update_state_on_changed(HOST_PNICS)
    def _translate_pnics(self, pnics):
        """Translate the host pnics data from vCenter."""

        row_data = VCenterDriver.convert_objs(pnics,
                                              VCenterDriver.pnic_translator)
        return row_data

    @ds_utils.update_state_on_changed(HOST_VNICS)
    def _translate_vnics(self, vnics):
        """Translate the host vnics data from vCenter."""

        row_data = VCenterDriver.convert_objs(vnics,
                                              VCenterDriver.vnic_translator)
        return row_data

    def _get_hosts_and_nics(self):
        """Convert vCenter host object to simple format.

        First the raw host data acquired from vCenter is parsed and
        organized into a simple format that can be read by congress
        translators. This creates three lists, hosts, pnics and vnics.
        These lists are then parsed by congress translators to create tables.
        """
        rawhosts = self._get_hosts_from_vcenter()
        hosts = []
        pnics = []
        vnics = []
        for host in rawhosts['objects']:
            h = {}
            h['vCenter_id'] = host.obj['value']
            for prop in host['propSet']:
                if prop.name == "hardware.systemInfo.uuid":
                    h['uuid'] = prop.val
                    break
            for prop in host['propSet']:
                if prop.name == "name":
                    h['name'] = prop.val
                    continue
                if prop.name == "config.network.dnsConfig.address":
                    try:
                        h[self.HOST_DNS] = prop.val.string
                    except AttributeError:
                        h[self.HOST_DNS] = ["No DNS IP adddresses configured"]
                    continue
                if prop.name == "config.network.pnic":
                    for pnic in prop.val.PhysicalNic:
                        p = {}
                        p['host_uuid'] = h['uuid']
                        p['mac'] = pnic['mac']
                        p['device'] = pnic['device']
                        p['ipAddress'] = pnic['spec']['ip']['ipAddress']
                        p['subnetMask'] = pnic['spec']['ip']['subnetMask']
                        pnics.append(p)
                if prop.name == "config.network.vnic":
                    for vnic in prop.val.HostVirtualNic:
                        v = {}
                        v['host_uuid'] = h['uuid']
                        v['device'] = vnic['device']
                        v['portgroup'] = vnic['portgroup']
                        v['mac'] = vnic['spec']['mac']
                        v['ipAddress'] = vnic['spec']['ip']['ipAddress']
                        v['subnetMask'] = vnic['spec']['ip']['subnetMask']
                        vnics.append(v)
            hosts.append(h)
        # cached the hosts for vms
        self.hosts = hosts
        return hosts, pnics, vnics

    @ds_utils.update_state_on_changed(VMS)
    def _translate_vms(self, vms):
        """Translate the VM data from vCenter."""

        row_data = VCenterDriver.convert_objs(vms,
                                              VCenterDriver.vms_translator)
        return row_data

    def _get_vms(self):
        rawvms = self._get_vms_from_vcenter()
        vms = []
        for vm in rawvms['objects']:
            v = {}
            for prop in vm['propSet']:
                if prop.name == "name":
                    v['name'] = prop.val
                    continue
                if prop.name == "config.uuid":
                    v['uuid'] = prop.val
                    continue
                if prop.name == "config.annotation":
                    v['annotation'] = prop.val
                    continue
                if prop.name == "summary.config.vmPathName":
                    v['pathName'] = prop.val
                    continue
                if prop.name == "summary.config.memorySizeMB":
                    v['memorySizeMB'] = prop.val
                    continue
                if prop.name == "summary.quickStats":
                    v['MemoryUsage'] = prop.val['guestMemoryUsage']
                    v['CpuDemand'] = prop.val['overallCpuDemand']
                    v['CpuUsage'] = prop.val['overallCpuUsage']
                    continue
                if prop.name == "summary.overallStatus":
                    v['status'] = prop.val
                if prop.name == "summary.storage":
                    v['committedStorage'] = prop.val['committed']
                    v['uncommittedStorage'] = prop.val['uncommitted']
                    continue
                if prop.name == 'runtime.host':
                    for host in self.hosts:
                        if host['vCenter_id'] == prop.val['value']:
                            v['host_uuid'] = host['uuid']
                            continue
                        continue
            vms.append(v)
        return vms

    def _get_hosts_from_vcenter(self):
        """Called to pull host data from vCenter

        """

        dataFields = ['name',
                      'hardware.systemInfo.uuid',
                      'config.network.dnsConfig.address',
                      'config.network.pnic',
                      'config.network.vnic']
        return self.session.invoke_api(vim_util, 'get_objects',
                                       self.session.vim, 'HostSystem',
                                       self.max_Hosts, dataFields)

    def _get_vms_from_vcenter(self):
        """Called to pull VM data from vCenter

        """

        dataFields = ['name',
                      'config.uuid',
                      'config.annotation',
                      'summary.config.vmPathName',
                      'runtime.host',
                      'summary.config.memorySizeMB',
                      'summary.quickStats',
                      'summary.overallStatus',
                      'summary.storage']
        return self.session.invoke_api(vim_util, 'get_objects',
                                       self.session.vim, 'VirtualMachine',
                                       self.max_VMs, dataFields)

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.session, action, action_args)



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
from oslo_log import log as logging
from oslo_middleware import request_id
import webob.dec
import webob.exc

from congress.common import config
from congress.common import wsgi
from congress import context


LOG = logging.getLogger(__name__)


class CongressKeystoneContext(wsgi.Middleware):
    """Make a request context from keystone headers."""

    @webob.dec.wsgify
    def __call__(self, req):
        # Determine the user ID
        user_id = req.headers.get('X_USER_ID')
        if not user_id:
            LOG.debug(_("X_USER_ID is not found in request"))
            return webob.exc.HTTPUnauthorized()

        # Determine the tenant
        tenant_id = req.headers.get('X_PROJECT_ID')

        # Suck out the roles
        roles = [r.strip() for r in req.headers.get('X_ROLES', '').split(',')]

        # Human-friendly names
        tenant_name = req.headers.get('X_PROJECT_NAME')
        user_name = req.headers.get('X_USER_NAME')

        # Use request_id if already set
        req_id = req.environ.get(request_id.ENV_REQUEST_ID)

        # Create a context with the authentication data
        ctx = context.RequestContext(user_id, tenant_id, roles=roles,
                                     user_name=user_name,
                                     tenant_name=tenant_name,
                                     request_id=req_id)

        # Inject the context...
        req.environ['congress.context'] = ctx

        return self.application


def pipeline_factory(loader, global_conf, **local_conf):
    """Create a paste pipeline based on the 'auth_strategy' config option."""
    config.set_config_defaults()
    pipeline = local_conf[cfg.CONF.auth_strategy]
    pipeline = pipeline.split()
    filters = [loader.get_filter(n) for n in pipeline[:-1]]
    app = loader.get_app(pipeline[-1])
    filters.reverse()
    for filter in filters:
        app = filter(app)
    return app



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import uuid

from oslo_config import cfg

from congress.api import status_model as st_model
from congress.api import webservice
from congress import harness
from congress.managers import datasource as datasource_manager
from congress.tests import base
from congress.tests import helper


class TestStatusModel(base.SqlTestCase):
    def setUp(self):
        super(TestStatusModel, self).setUp()
        # Here we load the fake driver
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        # NOTE(arosen): this set of tests, tests to deeply. We don't have
        # any tests currently testing cage. Once we do we should mock out
        # cage so we don't have to create one here.

        self.cage = harness.create(helper.root_path())
        self.ds_mgr = datasource_manager.DataSourceManager
        self.ds_mgr.validate_configured_drivers()
        req = {'driver': 'fake_datasource',
               'name': 'fake_datasource'}
        req['config'] = {'auth_url': 'foo',
                         'username': 'foo',
                         'password': 'password',
                         'tenant_name': 'foo'}
        self.datasource = self.ds_mgr.add_datasource(req)
        engine = self.cage.service_object('engine')
        self.status_model = st_model.StatusModel("status_schema", {},
                                                 policy_engine=engine,
                                                 datasource_mgr=self.ds_mgr)

    def test_get_datasource_status(self):
        context = {'ds_id': self.datasource['id']}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status_keys = ['last_updated', 'subscriptions',
                                'last_error', 'subscribers',
                                'initialized', 'number_of_updates']
        self.assertEqual(set(expected_status_keys), set(status.keys()))

    def test_get_invalid_datasource_status(self):
        context = {'ds_id': 'invalid_id'}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_policy_id_status(self):
        policy_model = self.cage.getservice(name='api-policy')['object']
        result = policy_model.add_item({'name': 'test_policy'}, {})

        context = {'policy_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_policy',
                           'id': result[0]}
        self.assertEqual(expected_status, status)

        # test with policy_name
        context = {'policy_id': result[1]['name']}
        status = self.status_model.get_item(None, {}, context=context)
        self.assertEqual(expected_status, status)

    def test_invalid_policy_id_status(self):
        invalid_id = uuid.uuid4()
        context = {'policy_id': invalid_id}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_rule_status_policy_id(self):
        policy_model = self.cage.getservice(name='api-policy')['object']
        result = policy_model.add_item({'name': 'test_policy'}, {})
        policy_id = result[0]
        policy_name = result[1]['name']

        rule_model = self.cage.getservice(name='api-rule')['object']
        result = rule_model.add_item({'name': 'test_rule',
                                      'rule': 'p(x) :- q(x)'}, {},
                                     context={'policy_id': 'test_policy'})

        context = {'policy_id': policy_id, 'rule_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_rule',
                           'id': result[0],
                           'comment': '',
                           'original_str': 'p(x) :- q(x)'}
        self.assertEqual(expected_status, status)

        # test with policy_name
        context = {'policy_id': policy_name, 'rule_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_rule',
                           'id': result[0],
                           'comment': '',
                           'original_str': 'p(x) :- q(x)'}
        self.assertEqual(expected_status, status)

    def test_rule_status_invalid_rule_policy_id(self):
        policy_model = self.cage.getservice(name='api-policy')['object']
        result = policy_model.add_item({'name': 'test_policy'}, {})
        policy_id = result[0]
        invalid_rule = uuid.uuid4()

        context = {'policy_id': policy_id, 'rule_id': invalid_rule}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_rule_status_invalid_policy_id(self):
        invalid_policy = uuid.uuid4()
        invalid_rule = uuid.uuid4()

        context = {'policy_id': invalid_policy, 'rule_id': invalid_rule}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)



# Copyright 2014 Plexxi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import six


class Node(object):
    def __init__(self, rPath=[], results=set()):
        self.destinations = set()

        self.results = results
        self.children = {}
        self.rPath = rPath

    def _remove(self, patternList, destination):

        word = patternList[0]

        if word in self.children:

            if len(patternList) == 1:
                if destination in self.children[word].destinations:
                    self.children[word].destinations.remove(destination)

                    if (len(self.children[word].destinations) == 0 and
                            len(self.children[word].children) == 0):

                        del self.children[word]

            else:
                self.children[word]._remove(patternList[1:], destination)

                if (len(self.children[word].destinations) == 0 and
                        len(self.children[word].children) == 0):

                    del self.children[word]

    def _add(self, patternList, destination):

        word = patternList[0]

        if word not in self.children:

            if word == "#":
                self.children['#'] = hashNode(
                    rPath=self.rPath + ['#'],
                    results=self.results)

            else:
                self.children[word] = Node(
                    rPath=self.rPath + [word],
                    results=self.results)

        if len(patternList) == 1:
            self.children[word].destinations.add(destination)

        else:
            self.children[word]._add(patternList[1:], destination)

    def update_results(self):
        if '#' in self.children:
            self.children['#'].update_results()

        self.results.update(self.destinations)

    def _lookup(self, keyList):
        word = keyList[0]

        if len(keyList) == 1:
            if word in self.children:
                self.children[word].update_results()

            if '*' in self.children:
                if word:
                    self.children['*'].update_results()

        else:
            if word in self.children:
                self.children[word]._lookup(keyList[1:])

            if '*' in self.children:
                if word:
                    self.children['*']._lookup(keyList[1:])

        if '#' in self.children:
            self.children['#']._lookup(keyList[:])
            self.children['#'].update_results()


class hashNode(Node):
    def _lookup(self, keyList):
        for i in range(len(keyList)):
            if keyList[i] in self.children:
                self.children[keyList[i]]._lookup(keyList[i:])

            if '*' in self.children:
                if keyList[i]:
                    self.children['*']._lookup(keyList[i:])

            if '#' in self.children:
                self.children['#']._lookup(keyList[i:])

        if keyList[-1] in self.children:
            self.children[keyList[-1]].update_results()

        if '*' in self.children:
            if keyList[-1]:
                self.children['*'].update_results()


class routeTable(Node):
    def add(self, pattern, destination):
        if type(pattern) == list:
            for p in pattern:
                wordList = p.split('.')
                self._add(wordList, destination)
        elif isinstance(pattern, six.string_types):
            wordList = pattern.split('.')
            self._add(wordList, destination)

    def remove(self, pattern, destination):
        if type(pattern) == list:
            for p in pattern:
                wordList = p.split('.')
                self._remove(wordList, destination)
        elif isinstance(pattern, six.string_types):
            wordList = pattern.split('.')
            self._remove(wordList, destination)

    def lookup(self, key):
        self.results.clear()

        wordList = key.split('.')

        self._lookup(wordList)

        return self.results
